# SPDX-License-Identifier: GPL-2.0-only
# This file is part of Scapy
# See https://scapy.net/ for more information
# Copyright (C) 2012 Luca Invernizzi <invernizzi.l@gmail.com>
# Copyright (C) 2012 Steeve Barbeau <http://www.sbarbeau.fr>
# Copyright (C) 2019 Gabriel Potter <gabriel[]potter[]fr>

"""
HTTP 1.0 layer.

Load using::

    from scapy.layers.http import *

Or (console only)::

    >>> load_layer("http")

Note that this layer ISN'T loaded by default, as quite experimental for now.

To follow HTTP packets streams = group packets together to get the
whole request/answer, use ``TCPSession`` as::

    >>> sniff(session=TCPSession)  # Live on-the-flow session
    >>> sniff(offline="./http_chunk.pcap", session=TCPSession)  # pcap

This will decode HTTP packets using ``Content_Length`` or chunks,
and will also decompress the packets when needed.
Note: on failure, decompression will be ignored.

You can turn auto-decompression/auto-compression off with::

    >>> conf.contribs["http"]["auto_compression"] = False

(Defaults to True)

You can also turn auto-chunking/dechunking off with::

    >>> conf.contribs["http"]["auto_chunk"] = False

(Defaults to True)
"""

# This file is a rewritten version of the former scapy_http plugin.
# It was reimplemented for scapy 2.4.3+ using sessions, stream handling.
# Original Authors : Steeve Barbeau, Luca Invernizzi

import base64
import datetime
import gzip
import io
import os
import re
import socket
import ssl
import struct
import subprocess

from enum import Enum

from scapy.compat import plain_str, bytes_encode

from scapy.automaton import Automaton, ATMT
from scapy.config import conf
from scapy.consts import WINDOWS
from scapy.error import warning, log_loading, log_interactive, Scapy_Exception
from scapy.fields import StrField
from scapy.packet import Packet, bind_layers, bind_bottom_up, Raw
from scapy.supersocket import StreamSocket, SSLStreamSocket
from scapy.utils import get_temp_file, ContextManagerSubprocess

from scapy.layers.gssapi import (
    ChannelBindingType,
    GSSAPI_BLOB,
    GSS_C_NO_CHANNEL_BINDINGS,
    GSS_S_COMPLETE,
    GSS_S_CONTINUE_NEEDED,
    GSS_S_FAILURE,
    GSS_S_FLAGS,
    GssChannelBindings,
)
from scapy.layers.inet import TCP

try:
    import brotli

    _is_brotli_available = True
except ImportError:
    _is_brotli_available = False

try:
    import lzw

    _is_lzw_available = True
except ImportError:
    _is_lzw_available = False

try:
    import zstandard

    _is_zstd_available = True
except ImportError:
    _is_zstd_available = False

if "http" not in conf.contribs:
    conf.contribs["http"] = {}
    conf.contribs["http"]["auto_compression"] = True
    conf.contribs["http"]["auto_chunk"] = True

# https://en.wikipedia.org/wiki/List_of_HTTP_header_fields

GENERAL_HEADERS = [
    "Cache-Control",
    "Connection",
    "Permanent",
    "Content-Length",
    "Content-MD5",
    "Content-Type",
    "Date",
    "Keep-Alive",
    "Pragma",
    "Upgrade",
    "Via",
    "Warning",
]

COMMON_UNSTANDARD_GENERAL_HEADERS = ["X-Request-ID", "X-Correlation-ID"]

REQUEST_HEADERS = [
    "A-IM",
    "Accept",
    "Accept-Charset",
    "Accept-Encoding",
    "Accept-Language",
    "Accept-Datetime",
    "Access-Control-Request-Method",
    "Access-Control-Request-Headers",
    "Authorization",
    "Cookie",
    "Expect",
    "Forwarded",
    "From",
    "Host",
    "HTTP2-Settings",
    "If-Match",
    "If-Modified-Since",
    "If-None-Match",
    "If-Range",
    "If-Unmodified-Since",
    "Max-Forwards",
    "Origin",
    "Proxy-Authorization",
    "Range",
    "Referer",
    "TE",
    "User-Agent",
]

COMMON_UNSTANDARD_REQUEST_HEADERS = [
    "Upgrade-Insecure-Requests",
    "X-Requested-With",
    "DNT",
    "X-Forwarded-For",
    "X-Forwarded-Host",
    "X-Forwarded-Proto",
    "Front-End-Https",
    "X-Http-Method-Override",
    "X-ATT-DeviceId",
    "X-Wap-Profile",
    "Proxy-Connection",
    "X-UIDH",
    "X-Csrf-Token",
    "Save-Data",
]

RESPONSE_HEADERS = [
    "Access-Control-Allow-Origin",
    "Access-Control-Allow-Credentials",
    "Access-Control-Expose-Headers",
    "Access-Control-Max-Age",
    "Access-Control-Allow-Methods",
    "Access-Control-Allow-Headers",
    "Accept-Patch",
    "Accept-Ranges",
    "Age",
    "Allow",
    "Alt-Svc",
    "Content-Disposition",
    "Content-Encoding",
    "Content-Language",
    "Content-Location",
    "Content-Range",
    "Delta-Base",
    "ETag",
    "Expires",
    "IM",
    "Last-Modified",
    "Link",
    "Location",
    "P3P",
    "Proxy-Authenticate",
    "Public-Key-Pins",
    "Retry-After",
    "Server",
    "Set-Cookie",
    "Strict-Transport-Security",
    "Trailer",
    "Transfer-Encoding",
    "Tk",
    "Vary",
    "WWW-Authenticate",
    "X-Frame-Options",
]

COMMON_UNSTANDARD_RESPONSE_HEADERS = [
    "Content-Security-Policy",
    "X-Content-Security-Policy",
    "X-WebKit-CSP",
    "Refresh",
    "Status",
    "Timing-Allow-Origin",
    "X-Content-Duration",
    "X-Content-Type-Options",
    "X-Powered-By",
    "X-UA-Compatible",
    "X-XSS-Protection",
]

# Dissection / Build tools


def _strip_header_name(name):
    """Takes a header key (i.e., "Host" in "Host: www.google.com",
    and returns a stripped representation of it
    """
    return plain_str(name.strip()).replace("-", "_")


def _header_line(name, val):
    """Creates a HTTP header line"""
    # Python 3.4 doesn't support % on bytes
    return bytes_encode(name) + b": " + bytes_encode(val)


def _parse_headers(s):
    headers = s.split(b"\r\n")
    headers_found = {}
    for header_line in headers:
        try:
            key, value = header_line.split(b":", 1)
        except ValueError:
            continue
        header_key = _strip_header_name(key).lower()
        headers_found[header_key] = (key, value.strip())
    return headers_found


def _parse_headers_and_body(s):
    """Takes a HTTP packet, and returns a tuple containing:
    _ the first line (e.g., "GET ...")
    _ the headers in a dictionary
    _ the body
    """
    crlfcrlf = b"\r\n\r\n"
    crlfcrlfIndex = s.find(crlfcrlf)
    if crlfcrlfIndex != -1:
        headers = s[: crlfcrlfIndex + len(crlfcrlf)]
        body = s[crlfcrlfIndex + len(crlfcrlf) :]
    else:
        headers = s
        body = b""
    first_line, headers = headers.split(b"\r\n", 1)
    return first_line.strip(), _parse_headers(headers), body


def _dissect_headers(obj, s):
    """Takes a HTTP packet as the string s, and populates the scapy layer obj
    (either HTTPResponse or HTTPRequest). Returns the first line of the
    HTTP packet, and the body
    """
    first_line, headers, body = _parse_headers_and_body(s)
    for f in obj.fields_desc:
        # We want to still parse wrongly capitalized fields
        stripped_name = _strip_header_name(f.name).lower()
        try:
            _, value = headers.pop(stripped_name)
        except KeyError:
            continue
        obj.setfieldval(f.name, value)
    if headers:
        headers = dict(headers.values())
        obj.setfieldval("Unknown_Headers", headers)
    return first_line, body


class _HTTPContent(Packet):
    __slots__ = ["_original_len"]

    # https://developer.mozilla.org/fr/docs/Web/HTTP/Headers/Transfer-Encoding
    def _get_encodings(self):
        encodings = []
        if isinstance(self, HTTPResponse):
            if self.Transfer_Encoding:
                encodings += [
                    plain_str(x).strip().lower()
                    for x in plain_str(self.Transfer_Encoding).split(",")
                ]
            if self.Content_Encoding:
                encodings += [
                    plain_str(x).strip().lower()
                    for x in plain_str(self.Content_Encoding).split(",")
                ]
        return encodings

    def hashret(self):
        return b"HTTP1"

    def post_dissect(self, s):
        self._original_len = len(s)
        encodings = self._get_encodings()
        # Un-chunkify
        if conf.contribs["http"]["auto_chunk"] and "chunked" in encodings:
            data = b""
            while s:
                length, _, body = s.partition(b"\r\n")
                try:
                    length = int(length, 16)
                except ValueError:
                    # Not a valid chunk. Ignore
                    break
                else:
                    load = body[:length]
                    if body[length : length + 2] != b"\r\n":
                        # Invalid chunk. Ignore
                        break
                    s = body[length + 2 :]
                    data += load
            if not s:
                s = data
        if not conf.contribs["http"]["auto_compression"]:
            return s
        # Decompress
        try:
            if "deflate" in encodings:
                import zlib

                s = zlib.decompress(s)
            elif "gzip" in encodings:
                s = gzip.decompress(s)
            elif "compress" in encodings:
                if _is_lzw_available:
                    s = lzw.decompress(s)
                else:
                    log_loading.info(
                        "Can't import lzw. compress decompression " "will be ignored !"
                    )
            elif "br" in encodings:
                if _is_brotli_available:
                    s = brotli.decompress(s)
                else:
                    log_loading.info(
                        "Can't import brotli. brotli decompression " "will be ignored !"
                    )
            elif "zstd" in encodings:
                if _is_zstd_available:
                    # Using its streaming API since its simple API could handle
                    # only cases where there is content size data embedded in
                    # the frame
                    bio = io.BytesIO(s)
                    reader = zstandard.ZstdDecompressor().stream_reader(bio)
                    s = reader.read()
                else:
                    log_loading.info(
                        "Can't import zstandard. zstd decompression "
                        "will be ignored !"
                    )
        except Exception:
            # Cannot decompress - probably incomplete data
            pass
        return s

    def post_build(self, pkt, pay):
        encodings = self._get_encodings()
        if conf.contribs["http"]["auto_compression"]:
            # Compress
            if "deflate" in encodings:
                import zlib

                pay = zlib.compress(pay)
            elif "gzip" in encodings:
                pay = gzip.compress(pay)
            elif "compress" in encodings:
                if _is_lzw_available:
                    pay = lzw.compress(pay)
                else:
                    log_loading.info(
                        "Can't import lzw. compress compression " "will be ignored !"
                    )
            elif "br" in encodings:
                if _is_brotli_available:
                    pay = brotli.compress(pay)
                else:
                    log_loading.info(
                        "Can't import brotli. brotli compression will " "be ignored !"
                    )
            elif "zstd" in encodings:
                if _is_zstd_available:
                    pay = zstandard.ZstdCompressor().compress(pay)
                else:
                    log_loading.info(
                        "Can't import zstandard. zstd compression will " "be ignored !"
                    )
        # Chunkify
        if conf.contribs["http"]["auto_chunk"] and "chunked" in encodings:
            # Dumb: 1 single chunk.
            pay = (b"%X" % len(pay)) + b"\r\n" + pay + b"\r\n0\r\n\r\n"
        return pkt + pay

    def self_build(self, **kwargs):
        """Takes an HTTPRequest or HTTPResponse object, and creates its
        string representation."""
        if not isinstance(self.underlayer, HTTP):
            warning("An HTTPResponse/HTTPRequest should always be below an HTTP")
        # Check for cache
        if self.raw_packet_cache is not None:
            return self.raw_packet_cache
        p = b""
        encodings = self._get_encodings()
        # Walk all the fields, in order
        for i, f in enumerate(self.fields_desc):
            if f.name == "Unknown_Headers":
                continue
            # Get the field value
            val = self.getfieldval(f.name)
            if not val:
                if f.name == "Content_Length" and "chunked" not in encodings:
                    # Add Content-Length anyways
                    val = str(len(self.payload or b""))
                elif f.name == "Date" and isinstance(self, HTTPResponse):
                    val = datetime.datetime.now(datetime.timezone.utc).strftime(
                        "%a, %d %b %Y %H:%M:%S GMT"
                    )
                else:
                    # Not specified. Skip
                    continue

            if i >= 3:
                val = _header_line(f.real_name, val)
            # Fields used in the first line have a space as a separator,
            # whereas headers are terminated by a new line
            if i <= 1:
                separator = b" "
            else:
                separator = b"\r\n"
            # Add the field into the packet
            p = f.addfield(self, p, val + separator)
        # Handle Unknown_Headers
        if self.Unknown_Headers:
            headers_text = b""
            for name, value in self.Unknown_Headers.items():
                headers_text += _header_line(name, value) + b"\r\n"
            p = self.get_field("Unknown_Headers").addfield(self, p, headers_text)
        # The packet might be empty, and in that case it should stay empty.
        if p:
            # Add an additional line after the last header
            p = f.addfield(self, p, b"\r\n")
        return p

    def guess_payload_class(self, payload):
        """Detect potential payloads"""
        if not hasattr(self, "Connection"):
            return super(_HTTPContent, self).guess_payload_class(payload)
        if self.Connection and b"Upgrade" in self.Connection:
            from scapy.contrib.http2 import H2Frame

            return H2Frame
        return super(_HTTPContent, self).guess_payload_class(payload)


class _HTTPHeaderField(StrField):
    """Modified StrField to handle HTTP Header names"""

    __slots__ = ["real_name"]

    def __init__(self, name, default):
        self.real_name = name
        name = _strip_header_name(name)
        StrField.__init__(self, name, default, fmt="H")

    def i2repr(self, pkt, x):
        if isinstance(x, bytes):
            return x.decode(errors="backslashreplace")
        return x


def _generate_headers(*args):
    """Generate the header fields based on their name"""
    # Order headers
    all_headers = []
    for headers in args:
        all_headers += headers
    # Generate header fields
    results = []
    for h in sorted(all_headers):
        results.append(_HTTPHeaderField(h, None))
    return results


# Create Request and Response packets


class HTTPRequest(_HTTPContent):
    name = "HTTP Request"
    fields_desc = (
        [
            # First line
            _HTTPHeaderField("Method", "GET"),
            _HTTPHeaderField("Path", "/"),
            _HTTPHeaderField("Http-Version", "HTTP/1.1"),
            # Headers
        ]
        + (
            _generate_headers(
                GENERAL_HEADERS,
                REQUEST_HEADERS,
                COMMON_UNSTANDARD_GENERAL_HEADERS,
                COMMON_UNSTANDARD_REQUEST_HEADERS,
            )
        )
        + [
            _HTTPHeaderField("Unknown-Headers", None),
        ]
    )

    def do_dissect(self, s):
        """From the HTTP packet string, populate the scapy object"""
        first_line, body = _dissect_headers(self, s)
        try:
            Method, Path, HTTPVersion = re.split(rb"\s+", first_line, maxsplit=2)
            self.setfieldval("Method", Method)
            self.setfieldval("Path", Path)
            self.setfieldval("Http_Version", HTTPVersion)
        except ValueError:
            pass
        if body:
            self.raw_packet_cache = s[: -len(body)]
        else:
            self.raw_packet_cache = s
        return body

    def mysummary(self):
        return self.sprintf("%HTTPRequest.Method% '%HTTPRequest.Path%' ")


class HTTPResponse(_HTTPContent):
    name = "HTTP Response"
    fields_desc = (
        [
            # First line
            _HTTPHeaderField("Http-Version", "HTTP/1.1"),
            _HTTPHeaderField("Status-Code", "200"),
            _HTTPHeaderField("Reason-Phrase", "OK"),
            # Headers
        ]
        + (
            _generate_headers(
                GENERAL_HEADERS,
                RESPONSE_HEADERS,
                COMMON_UNSTANDARD_GENERAL_HEADERS,
                COMMON_UNSTANDARD_RESPONSE_HEADERS,
            )
        )
        + [
            _HTTPHeaderField("Unknown-Headers", None),
        ]
    )

    def answers(self, other):
        return HTTPRequest in other

    def do_dissect(self, s):
        """From the HTTP packet string, populate the scapy object"""
        first_line, body = _dissect_headers(self, s)
        try:
            HTTPVersion, Status, Reason = re.split(rb"\s+", first_line, maxsplit=2)
            self.setfieldval("Http_Version", HTTPVersion)
            self.setfieldval("Status_Code", Status)
            self.setfieldval("Reason_Phrase", Reason)
        except ValueError:
            pass
        if body:
            self.raw_packet_cache = s[: -len(body)]
        else:
            self.raw_packet_cache = s
        return body

    def mysummary(self):
        return self.sprintf("%HTTPResponse.Status_Code% %HTTPResponse.Reason_Phrase%")


# General HTTP class + defragmentation


class HTTP(Packet):
    name = "HTTP 1"
    fields_desc = []
    show_indent = 0
    clsreq = HTTPRequest
    clsresp = HTTPResponse
    hdr = b"HTTP"
    reqmethods = b"|".join(
        [
            b"OPTIONS",
            b"GET",
            b"HEAD",
            b"POST",
            b"PUT",
            b"DELETE",
            b"TRACE",
            b"CONNECT",
        ]
    )

    @classmethod
    def dispatch_hook(cls, _pkt=None, *args, **kargs):
        if _pkt and len(_pkt) >= 9:
            from scapy.contrib.http2 import _HTTP2_types, H2Frame

            # To detect a valid HTTP2, we check that the type is correct
            # that the Reserved bit is set and length makes sense.
            while _pkt:
                if len(_pkt) < 9:
                    # Invalid total length
                    return cls
                if ord(_pkt[3:4]) not in _HTTP2_types:
                    # Invalid type
                    return cls
                length = struct.unpack("!I", b"\0" + _pkt[:3])[0] + 9
                if length > len(_pkt):
                    # Invalid length
                    return cls
                sid = struct.unpack("!I", _pkt[5:9])[0]
                if sid >> 31 != 0:
                    # Invalid Reserved bit
                    return cls
                _pkt = _pkt[length:]
            return H2Frame
        return cls

    # tcp_reassemble is used by TCPSession in session.py
    @classmethod
    def tcp_reassemble(cls, data, metadata, _):
        detect_end = metadata.get("detect_end", None)
        is_unknown = metadata.get("detect_unknown", True)
        # General idea of the following is explained at
        # https://datatracker.ietf.org/doc/html/rfc2616#section-4.4
        if not detect_end or is_unknown:
            metadata["detect_unknown"] = False
            http_packet = cls(data)
            # Detect packing method
            if not isinstance(http_packet.payload, _HTTPContent):
                return http_packet
            is_response = isinstance(http_packet.payload, cls.clsresp)
            # Packets may have a Content-Length we must honnor
            length = http_packet.Content_Length
            if length:
                # Parse the length as an integer
                try:
                    length = int(length)
                except ValueError:
                    length = None
            if length is not None:
                # The packet provides a Content-Length attribute: let's
                # use it. When the total size of the frags is high enough,
                # we have the packet

                # Subtract the length of the "HTTP*" layer
                if http_packet.payload.payload or length == 0:
                    http_length = len(data) - http_packet.payload._original_len
                    detect_end = lambda dat: len(dat) - http_length >= length
                else:
                    # The HTTP layer isn't fully received.
                    if metadata.get("tcp_end", False):
                        # This was likely a HEAD response. Ugh
                        detect_end = lambda dat: True
                    else:
                        detect_end = lambda dat: False
                        metadata["detect_unknown"] = True
            else:
                # It's not Content-Length based. It could be chunked
                encodings = http_packet[cls].payload._get_encodings()
                chunked = "chunked" in encodings
                if chunked:
                    detect_end = lambda dat: dat.endswith(b"0\r\n\r\n")
                # HTTP Requests that do not have any content,
                # end with a double CRLF. Same for HEAD responses
                elif isinstance(http_packet.payload, cls.clsreq):
                    detect_end = lambda dat: dat.endswith(b"\r\n\r\n")
                    # In case we are handling a HTTP Request,
                    # we want to continue assessing the data,
                    # to handle requests with a body (POST)
                    metadata["detect_unknown"] = True
                elif is_response and http_packet.Status_Code == b"101":
                    # If it's an upgrade response, it may also hold a
                    # different protocol data.
                    # make sure all headers are present
                    detect_end = lambda dat: dat.find(b"\r\n\r\n")
                else:
                    # If neither Content-Length nor chunked is specified,
                    # it means it's the TCP packet that contains the data,
                    # or that the information hasn't been given yet.
                    detect_end = lambda dat: metadata.get("tcp_end", False)
                    metadata["detect_unknown"] = True
            metadata["detect_end"] = detect_end
            if detect_end(data):
                return http_packet
        else:
            if detect_end(data):
                http_packet = cls(data)
                return http_packet

    def guess_payload_class(self, payload):
        """Decides if the payload is an HTTP Request or Response, or
        something else.
        """
        try:
            prog = re.compile(
                rb"^(?:"
                + self.reqmethods
                + rb") "
                + rb"(?:.+?) "
                + self.hdr
                + rb"/\d\.\d$"
            )
            crlfIndex = payload.index(b"\r\n")
            req = payload[:crlfIndex]
            result = prog.match(req)
            if result:
                return self.clsreq
            else:
                prog = re.compile(b"^" + self.hdr + rb"/\d\.\d \d\d\d .*$")
                result = prog.match(req)
                if result:
                    return self.clsresp
        except ValueError:
            # Anything that isn't HTTP but on port 80
            pass
        return Raw


class HTTP_AUTH_MECHS(Enum):
    NONE = "NONE"
    BASIC = "Basic"
    NTLM = "NTLM"
    NEGOTIATE = "Negotiate"


class HTTP_Client(object):
    """
    A basic HTTP client

    :param mech: one of HTTP_AUTH_MECHS
    :param ssl: whether to use HTTPS or not
    :param ssp: the SSP object to use for binding
    :param no_check_certificate: with SSL, do not check the certificate
    :param no_chan_bindings: force disable sending the channel bindings
    """

    def __init__(
        self,
        mech=HTTP_AUTH_MECHS.NONE,
        verb=True,
        sslcontext=None,
        ssp=None,
        no_check_certificate=False,
        no_chan_bindings=False,
    ):
        self.sock = None
        self._sockinfo = None
        self.authmethod = mech
        self.verb = verb
        self.sslcontext = sslcontext
        self.ssp = ssp
        self.sspcontext = None
        self.no_check_certificate = no_check_certificate
        self.no_chan_bindings = no_chan_bindings
        self.chan_bindings = GSS_C_NO_CHANNEL_BINDINGS

    def _connect_or_reuse(self, host, port=None, tls=False, timeout=5):
        # Get the port
        if port is None:
            port = 443 if tls else 80
        # If the current socket matches, keep it.
        if self._sockinfo == (host, port):
            return
        # A new socket is needed
        if self._sockinfo:
            self.close()
        sock = socket.socket()
        sock.setsockopt(socket.SOL_SOCKET, socket.SO_KEEPALIVE, 1)
        sock.settimeout(timeout)
        if self.verb:
            print(
                "\u2503 Connecting to %s on port %s%s..."
                % (
                    host,
                    port,
                    " with SSL" if tls else "",
                )
            )
        sock.connect((host, port))
        if self.verb:
            print(
                conf.color_theme.green(
                    "\u2514 Connected from %s" % repr(sock.getsockname())
                )
            )
        if tls:
            if self.sslcontext is None:
                if self.no_check_certificate:
                    context = ssl.SSLContext(ssl.PROTOCOL_TLS_CLIENT)
                    context.check_hostname = False
                    context.verify_mode = ssl.CERT_NONE
                else:
                    context = ssl.create_default_context()
            else:
                context = self.sslcontext
            sock = context.wrap_socket(sock, server_hostname=host)
            if self.ssp and not self.no_chan_bindings:
                # Compute the channel binding token (CBT)
                self.chan_bindings = GssChannelBindings.fromssl(
                    ChannelBindingType.TLS_SERVER_END_POINT,
                    sslsock=sock,
                )
            self.sock = SSLStreamSocket(sock, HTTP)
        else:
            self.sock = StreamSocket(sock, HTTP)
        # Store information regarding the current socket
        self._sockinfo = (host, port)

    def sr1(self, req, **kwargs):
        if self.verb:
            print(conf.color_theme.opening(">> %s" % req.summary()))
        resp = self.sock.sr1(
            HTTP() / req,
            verbose=0,
            **kwargs,
        )
        if self.verb:
            print(conf.color_theme.success("<< %s" % (resp and resp.summary())))
        return resp

    def request(
        self,
        url,
        data=b"",
        timeout=5,
        follow_redirects=True,
        http_headers={},
        **headers,
    ):
        """
        Perform a HTTP(s) request.

        :param url: the full URL to connect to.
            e.g. https://google.com/test
        :param data: the data to send as payload
        :param follow_redirects: if True, request() will follow 302 return codes
        :param http_headers: if specified, overwrites the HTTP headers
            (except Host and Path).
        :param headers: any additional HTTPRequest parameter to add.
            e.g. Method="POST"
        """
        # Parse request url
        m = re.match(r"(https?)://([^/:]+)(?:\:(\d+))?(/.*)?", url)
        if not m:
            raise ValueError("Bad URL !")
        transport, host, port, path = m.groups()
        if transport == "https":
            tls = True
        else:
            tls = False

        path = path or "/"
        port = port and int(port)

        # Connect (or reuse) socket
        self._connect_or_reuse(host, port=port, tls=tls, timeout=timeout)

        # Build request
        if ((tls and port != 443) or
                (not tls and port != 80)):
            host_hdr = "%s:%d" % (host, port)
        else:
            host_hdr = host

        headers.setdefault("Host", host_hdr)
        headers.setdefault("Path", path)

        if not http_headers:
            http_headers = {
                "Accept_Encoding": b"gzip, deflate",
                "Cache_Control": b"no-cache",
                "Pragma": b"no-cache",
                "Connection": b"keep-alive",
            }
        else:
            http_headers = {k.replace("-", "_"): v for k, v in http_headers.items()}
        http_headers.update(headers)
        req = HTTP() / HTTPRequest(**http_headers)
        if data:
            req /= data

        while True:
            # Perform the request.
            try:
                resp = self.sr1(req, timeout=timeout)
            except Exception:
                # Socket has died, restart.
                self._sockinfo = None
                self._connect_or_reuse(host, port=port, tls=tls, timeout=timeout)
                continue
            if not resp:
                break
            # First case: auth was required. Handle that
            if resp.Status_Code in [b"401", b"407"]:
                # Authentication required
                if self.authmethod in [
                    HTTP_AUTH_MECHS.NTLM,
                    HTTP_AUTH_MECHS.NEGOTIATE,
                ]:
                    # Parse authenticate
                    if b" " in resp.WWW_Authenticate:
                        method, data = resp.WWW_Authenticate.split(b" ", 1)
                        try:
                            ssp_blob = GSSAPI_BLOB(base64.b64decode(data))
                        except Exception:
                            raise Scapy_Exception("Invalid WWW-Authenticate")
                    else:
                        method = resp.WWW_Authenticate
                        ssp_blob = None
                    if plain_str(method) != self.authmethod.value:
                        raise Scapy_Exception("Invalid WWW-Authenticate")
                    # SPNEGO / Kerberos / NTLM
                    self.sspcontext, token, status = self.ssp.GSS_Init_sec_context(
                        self.sspcontext,
                        input_token=ssp_blob,
                        target_name="http/" + host,
                        req_flags=0,
                        chan_bindings=self.chan_bindings,
                    )
                    if status not in [GSS_S_COMPLETE, GSS_S_CONTINUE_NEEDED]:
                        raise Scapy_Exception("Authentication failure")
                    req.Authorization = (
                        self.authmethod.value.encode()
                        + b" "
                        + base64.b64encode(bytes(token))
                    )
                    continue
            # Second case: follow redirection
            if resp.Status_Code in [b"301", b"302"] and follow_redirects:
                return self.request(
                    resp.Location.decode(),
                    data=data,
                    timeout=timeout,
                    follow_redirects=follow_redirects,
                    **headers,
                )
            break
        return resp

    def close(self):
        if self.verb:
            print("X Connection to server closed\n")
        self.sock.close()


def http_request(
    host, path="/", port=None, timeout=3, display=False, tls=False, verbose=0, **headers
):
    """
    Util to perform an HTTP request.

    :param host: the host to connect to
    :param path: the path of the request (default /)
    :param port: the port (default 80/443)
    :param timeout: timeout before None is returned
    :param display: display the result in the default browser (default False)
    :param iface: interface to use. Changing this turns on "raw"
    :param headers: any additional headers passed to the request

    :returns: the HTTPResponse packet
    """
    client = HTTP_Client(HTTP_AUTH_MECHS.NONE, verb=verbose)
    if port is None:
        port = 443 if tls else 80
    ans = client.request(
        "http%s://%s:%s%s" % (tls and "s" or "", host, port, path),
        timeout=timeout,
    )

    if ans:
        if display:
            if Raw not in ans:
                warning("No HTTP content returned. Cannot display")
                return ans
            # Write file
            file = get_temp_file(autoext=".html")
            with open(file, "wb") as fd:
                fd.write(ans.load)
            # Open browser
            if WINDOWS:
                os.startfile(file)
            else:
                with ContextManagerSubprocess(conf.prog.universal_open):
                    subprocess.Popen([conf.prog.universal_open, file])
        return ans


# Bindings


bind_bottom_up(TCP, HTTP, sport=80)
bind_bottom_up(TCP, HTTP, dport=80)
bind_layers(TCP, HTTP, sport=80, dport=80)

bind_bottom_up(TCP, HTTP, sport=8080)
bind_bottom_up(TCP, HTTP, dport=8080)


# Automatons


class HTTP_Server(Automaton):
    """
    HTTP server automaton

    :param ssp: the SSP to serve. If None, unauthenticated (or basic).
    :param mech: the HTTP_AUTH_MECHS to use (default: NONE)
    :param require_cbt: require Channel Bindings to be valid (default: False)
    :param cbt_cert: the path to the certificate used for channel bindings.
                     Useful if behind a reverse proxy. (default: None)

    Other parameters:

    :param BASIC_IDENTITIES: a dict that contains {"user": "password"} for Basic
                             authentication.
    :param BASIC_REALM: the basic realm.
    """

    pkt_cls = HTTP

    def __init__(
        self,
        mech=HTTP_AUTH_MECHS.NONE,
        verb=True,
        ssp=None,
        require_cbt: bool = False,
        cbt_cert: str = None,
        *args,
        **kwargs,
    ):
        self.verb = verb
        if "sock" not in kwargs:
            raise ValueError(
                "HTTP_Server cannot be started directly ! Use HTTP_Server.spawn"
            )
        self.ssp = ssp
        self.authmethod = mech.value
        self.sspcontext = None

        # CBT settings
        self.ssp_req_flags = GSS_S_FLAGS.GSS_S_ALLOW_MISSING_BINDINGS
        if require_cbt:
            self.ssp_req_flags &= ~GSS_S_FLAGS.GSS_S_ALLOW_MISSING_BINDINGS
        if cbt_cert:
            self.chan_bindings = GssChannelBindings.fromssl(
                ChannelBindingType.TLS_SERVER_END_POINT,
                certfile=cbt_cert,
            )
        else:
            self.chan_bindings = GSS_C_NO_CHANNEL_BINDINGS

        # Auth settings
        self.basic = False
        self.BASIC_IDENTITIES = kwargs.pop("BASIC_IDENTITIES", {})
        self.BASIC_REALM = kwargs.pop("BASIC_REALM", "default")
        if mech == HTTP_AUTH_MECHS.BASIC:
            if not self.BASIC_IDENTITIES:
                raise ValueError("Please provide 'BASIC_IDENTITIES' !")
            if ssp is not None:
                raise ValueError("Can't use 'BASIC_IDENTITIES' with 'ssp' !")
            self.basic = True
        elif mech == HTTP_AUTH_MECHS.NONE:
            if ssp is not None:
                raise ValueError("Cannot use ssp with mech=NONE !")
        # Initialize
        Automaton.__init__(self, *args, **kwargs)

    def send(self, resp):
        self.sock.send(HTTP() / resp)

    def vprint(self, s=""):
        """
        Verbose print (if enabled)
        """
        if self.verb:
            if conf.interactive:
                log_interactive.info("> %s", s)
            else:
                print("> %s" % s)

    @ATMT.state(initial=1)
    def BEGIN(self):
        self.authenticated = False
        self.sspcontext = None

    @ATMT.receive_condition(BEGIN, prio=1)
    def should_authenticate(self, pkt):
        if self.authmethod == HTTP_AUTH_MECHS.NONE.value:
            raise self.SERVE(pkt)
        else:
            raise self.AUTH(pkt)

    @ATMT.state()
    def AUTH(self, pkt=None):
        if pkt is None:
            return
        if HTTPRequest in pkt:
            self.vprint(pkt.summary())
            if pkt.Method == b"CONNECT":
                # HTTP tunnel (proxy)
                proxy = True
            else:
                # HTTP non-tunnel
                proxy = False
            # Get authorization
            if proxy:
                authorization = pkt.Proxy_Authorization
            else:
                authorization = pkt.Authorization
            if not authorization:
                # Initial ask.
                data = self.authmethod
                if self.basic:
                    data += " realm='%s'" % self.BASIC_REALM
                self._ask_authorization(proxy, data)
                return
            # Parse authorization
            method, data = authorization.split(b" ", 1)
            if plain_str(method) != self.authmethod:
                self.debug(3, "Bad auth method.")
                raise self.AUTH_ERROR(proxy)
            try:
                data = base64.b64decode(data)
            except Exception:
                self.debug(3, "Couldn't unpack base64 of auth.")
                raise self.AUTH_ERROR(proxy)
            # Now process the authorization
            if not self.basic:
                try:
                    ssp_blob = GSSAPI_BLOB(data)
                except Exception:
                    self.sspcontext = None
                    self._ask_authorization(proxy, self.authmethod)
                    self.debug(3, "Couldn't unpack GSSAPI_BLOB of auth.")
                    raise self.AUTH_ERROR(proxy)
                # And call the SSP
                self.sspcontext, tok, status = self.ssp.GSS_Accept_sec_context(
                    self.sspcontext,
                    ssp_blob,
                    req_flags=self.ssp_req_flags,
                    chan_bindings=self.chan_bindings,
                )
            else:
                # This is actually Basic authentication
                try:
                    next(
                        True
                        for k, v in self.BASIC_IDENTITIES.items()
                        if ("%s:%s" % (k, v)).encode() == data
                    )
                    tok, status = None, GSS_S_COMPLETE
                except StopIteration:
                    self.debug(3, "Basic authentication failed with 'unknown user'.")
                    tok, status = None, GSS_S_FAILURE
            # Send answer
            if status not in [GSS_S_COMPLETE, GSS_S_CONTINUE_NEEDED]:
                self.debug(3, "Authentication failed: %s." % status)
                raise self.AUTH_ERROR(proxy)
            elif status == GSS_S_CONTINUE_NEEDED:
                data = self.authmethod.encode()
                if tok:
                    data += b" " + base64.b64encode(bytes(tok))
                self._ask_authorization(proxy, data)
                raise self.AUTH()
            else:
                # Authenticated !
                self.authenticated = True
                self.vprint("AUTH OK")
                raise self.SERVE(pkt)

    @ATMT.state()
    def AUTH_ERROR(self, proxy):
        self.sspcontext = None
        self._ask_authorization(proxy, self.authmethod)
        self.vprint("AUTH ERROR")

    @ATMT.condition(AUTH_ERROR)
    def allow_reauth(self):
        raise self.AUTH()

    def _ask_authorization(self, proxy, data):
        if proxy:
            self.send(
                HTTPResponse(
                    Status_Code=b"407",
                    Reason_Phrase=b"Proxy Authentication Required",
                    Proxy_Authenticate=data,
                )
            )
        else:
            self.send(
                HTTPResponse(
                    Status_Code=b"401",
                    Reason_Phrase=b"Unauthorized",
                    WWW_Authenticate=data,
                )
            )

    @ATMT.receive_condition(AUTH, prio=1)
    def received_unauthenticated(self, pkt):
        raise self.AUTH(pkt)

    @ATMT.eof(AUTH)
    def auth_eof(self):
        raise self.CLOSED()

    @ATMT.state(error=1)
    def ERROR(self):
        self.send(
            HTTPResponse(
                Status_Code="400",
                Reason_Phrase="Bad Request",
            )
        )

    @ATMT.state(final=1)
    def CLOSED(self):
        self.vprint("CLOSED")

    # Serving

    @ATMT.state()
    def SERVE(self, pkt=None):
        if pkt is None:
            return
        answer = self.answer(pkt)
        if answer:
            self.send(answer)
            self.vprint("%s -> %s" % (pkt.summary(), answer.summary()))
        else:
            self.vprint("%s" % pkt.summary())

    @ATMT.eof(SERVE)
    def serve_eof(self):
        raise self.CLOSED()

    @ATMT.receive_condition(SERVE)
    def new_request(self, pkt):
        raise self.SERVE(pkt)

    # DEV: overwrite this function

    def answer(self, pkt):
        """
        HTTP_server answer function.

        :param pkt: a HTTPRequest packet
        :returns: a HTTPResponse packet
        """
        if pkt.Path == b"/":
            return HTTPResponse() / (
                "<!doctype html><html><body><h1>OK</h1></body></html>"
            )
        else:
            return HTTPResponse(
                Status_Code=b"404",
                Reason_Phrase=b"Not Found",
            ) / ("<!doctype html><html><body><h1>404 - Not Found</h1></body></html>")


class HTTPS_Server(HTTP_Server):
    """
    HTTPS server automaton

    This has the same arguments and attributes as HTTP_Server, with the addition of:

    :param sslcontext: an optional SSLContext object.
                       If used, key is ignored but cert can still be used for
                       channel bindings.
    :param cert: path to the certificate
    :param key: path to the key
    :param require_cbt: require Channel Bindings to be valid
    """

    socketcls = None

    def __init__(
        self,
        mech=HTTP_AUTH_MECHS.NONE,
        verb=True,
        cert=None,
        key=None,
        sslcontext=None,
        ssp=None,
        require_cbt=False,
        *args,
        **kwargs,
    ):
        if "sock" not in kwargs:
            raise ValueError(
                "HTTPS_Server cannot be started directly ! Use HTTPS_Server.spawn"
            )
        # wrap socket in SSL
        if sslcontext is None:
            context = ssl.SSLContext(ssl.PROTOCOL_TLS_SERVER)
            context.load_cert_chain(cert, key)
        else:
            context = sslcontext
        kwargs["sock"] = SSLStreamSocket(
            context.wrap_socket(kwargs["sock"], server_side=True),
            self.pkt_cls,
        )

        # Call super
        super(HTTPS_Server, self).__init__(
            mech=mech,
            verb=verb,
            ssp=ssp,
            cbt_cert=cert,
            require_cbt=require_cbt,
            *args,
            **kwargs,
        )



# SPDX-License-Identifier: GPL-2.0-only
# This file is part of Scapy
# See https://scapy.net/ for more information
# Copyright (C) Nils Weiss <nils@we155.de>

# scapy.contrib.description = ISO-TP (ISO 15765-2) Native Socket Library
# scapy.contrib.status = library

import ctypes
import socket
import struct
from ctypes.util import find_library
# Typing imports
from typing import (
    Any,
    Optional,
    Union,
    Tuple,
    Type,
    cast,
)

from scapy.arch.linux import get_last_packet_timestamp, SIOCGIFINDEX
from scapy.config import conf
from scapy.contrib.isotp import log_isotp
from scapy.contrib.isotp.isotp_packet import ISOTP
from scapy.data import SO_TIMESTAMPNS
from scapy.error import Scapy_Exception
from scapy.layers.can import CAN_MTU, CAN_FD_MTU, CAN_MAX_DLEN, CAN_FD_MAX_DLEN
from scapy.packet import Packet
from scapy.supersocket import SuperSocket

LIBC = ctypes.cdll.LoadLibrary(find_library("c"))  # type: ignore

CAN_ISOTP = 6  # ISO 15765-2 Transport Protocol

SOL_CAN_BASE = 100  # from can.h
SOL_CAN_ISOTP = SOL_CAN_BASE + CAN_ISOTP
# /* for socket options affecting the socket (not the global system) */
CAN_ISOTP_OPTS = 1  # /* pass struct can_isotp_options */
CAN_ISOTP_RECV_FC = 2  # /* pass struct can_isotp_fc_options */

# /* sockopts to force stmin timer values for protocol regression tests */
CAN_ISOTP_TX_STMIN = 3  # /* pass __u32 value in nano secs    */
CAN_ISOTP_RX_STMIN = 4  # /* pass __u32 value in nano secs   */
CAN_ISOTP_LL_OPTS = 5  # /* pass struct can_isotp_ll_options */

CAN_ISOTP_LISTEN_MODE = 0x001  # /* listen only (do not send FC) */
CAN_ISOTP_EXTEND_ADDR = 0x002  # /* enable extended addressing */
CAN_ISOTP_TX_PADDING = 0x004  # /* enable CAN frame padding tx path */
CAN_ISOTP_RX_PADDING = 0x008  # /* enable CAN frame padding rx path */
CAN_ISOTP_CHK_PAD_LEN = 0x010  # /* check received CAN frame padding */
CAN_ISOTP_CHK_PAD_DATA = 0x020  # /* check received CAN frame padding */
CAN_ISOTP_HALF_DUPLEX = 0x040  # /* half duplex error state handling */
CAN_ISOTP_FORCE_TXSTMIN = 0x080  # /* ignore stmin from received FC */
CAN_ISOTP_FORCE_RXSTMIN = 0x100  # /* ignore CFs depending on rx stmin */
CAN_ISOTP_RX_EXT_ADDR = 0x200  # /* different rx extended addressing */

# /* default values */
CAN_ISOTP_DEFAULT_FLAGS = 0
CAN_ISOTP_DEFAULT_EXT_ADDRESS = 0x00
CAN_ISOTP_DEFAULT_PAD_CONTENT = 0xCC  # /* prevent bit-stuffing */
CAN_ISOTP_DEFAULT_FRAME_TXTIME = 0
CAN_ISOTP_DEFAULT_RECV_BS = 0
CAN_ISOTP_DEFAULT_RECV_STMIN = 0x00
CAN_ISOTP_DEFAULT_RECV_WFTMAX = 0
CAN_ISOTP_DEFAULT_LL_MTU = CAN_MTU
CAN_ISOTP_CANFD_MTU = CAN_FD_MTU
CAN_ISOTP_DEFAULT_LL_TX_DL = CAN_MAX_DLEN
CAN_FD_ISOTP_DEFAULT_LL_TX_DL = CAN_FD_MAX_DLEN
CAN_ISOTP_DEFAULT_LL_TX_FLAGS = 0

CANFD_BRS = 1  # /* CAN FD Bit Rate Switch */
CANFD_ESI = 2  # /* CAN FD Error State Indicator */
CANFD_FDF = 4  # /* CAN FD FD Flag */


class tp(ctypes.Structure):
    # This struct is only used within the sockaddr_can struct
    _fields_ = [("rx_id", ctypes.c_uint32),
                ("tx_id", ctypes.c_uint32)]


class addr_info(ctypes.Union):
    # This struct is only used within the sockaddr_can struct
    # This union is to future proof for future can address information
    _fields_ = [("tp", tp)]


class sockaddr_can(ctypes.Structure):
    # See /usr/include/linux/can.h for original struct
    _fields_ = [("can_family", ctypes.c_uint16),
                ("can_ifindex", ctypes.c_int),
                ("can_addr", addr_info)]


class ifreq(ctypes.Structure):
    # The two fields in this struct were originally unions.
    # See /usr/include/net/if.h for original struct
    _fields_ = [("ifr_name", ctypes.c_char * 16),
                ("ifr_ifindex", ctypes.c_int)]


class ISOTPNativeSocket(SuperSocket):
    """
    ISOTPSocket using the can-isotp kernel module

    :param iface: a CANSocket instance or an interface name
    :param tx_id: the CAN identifier of the sent CAN frames
    :param rx_id: the CAN identifier of the received CAN frames
    :param ext_address: the extended address of the sent ISOTP frames
    :param rx_ext_address: the extended address of the received ISOTP frames
    :param bs: block size sent in Flow Control ISOTP frames
    :param stmin: minimum desired separation time sent in
                  Flow Control ISOTP frames
    :param padding: If True, pads sending packets with 0x00 which not
                    count to the payload.
                    Does not affect receiving packets.
    :param listen_only: Does not send Flow Control frames if a First Frame is
                        received
    :param frame_txtime: Separation time between two CAN frames during send
    :param basecls: base class of the packets emitted by this socket
    """
    desc = "read/write packets at a given CAN interface using CAN_ISOTP socket "  # noqa: E501
    can_isotp_options_fmt = "@2I4B"
    can_isotp_fc_options_fmt = "@3B"
    can_isotp_ll_options_fmt = "@3B"
    sockaddr_can_fmt = "@H3I"
    auxdata_available = True

    def __build_can_isotp_options(
            self,
            flags=CAN_ISOTP_DEFAULT_FLAGS,
            frame_txtime=CAN_ISOTP_DEFAULT_FRAME_TXTIME,
            ext_address=CAN_ISOTP_DEFAULT_EXT_ADDRESS,
            txpad_content=CAN_ISOTP_DEFAULT_PAD_CONTENT,
            rxpad_content=CAN_ISOTP_DEFAULT_PAD_CONTENT,
            rx_ext_address=CAN_ISOTP_DEFAULT_EXT_ADDRESS):
        # type: (int, int, int, int, int, int) -> bytes
        return struct.pack(self.can_isotp_options_fmt,
                           flags,
                           frame_txtime,
                           ext_address,
                           txpad_content,
                           rxpad_content,
                           rx_ext_address)

    # == Must use native not standard types for packing ==
    # struct can_isotp_options {
    # __u32 flags;            /* set flags for isotp behaviour.       */
    #                         /* __u32 value : flags see below        */
    #
    # __u32 frame_txtime;     /* frame transmission time (N_As/N_Ar)  */
    #                       /* __u32 value : time in nano secs      */
    #
    # __u8  ext_address;      /* set address for extended addressing  */
    #                         /* __u8 value : extended address        */
    #
    # __u8  txpad_content;    /* set content of padding byte (tx)     */
    #                         /* __u8 value : content on tx path      */
    #
    # __u8  rxpad_content;    /* set content of padding byte (rx)     */
    #                         /* __u8 value : content on rx path      */
    #
    # __u8  rx_ext_address;   /* set address for extended addressing  */
    #                         /* __u8 value : extended address (rx)   */
    # };

    def __build_can_isotp_fc_options(self,
                                     bs=CAN_ISOTP_DEFAULT_RECV_BS,
                                     stmin=CAN_ISOTP_DEFAULT_RECV_STMIN,
                                     wftmax=CAN_ISOTP_DEFAULT_RECV_WFTMAX):
        # type: (int, int, int) -> bytes
        return struct.pack(self.can_isotp_fc_options_fmt,
                           bs,
                           stmin,
                           wftmax)

    # == Must use native not standard types for packing ==
    # struct can_isotp_fc_options {
    #
    # __u8  bs;               /* blocksize provided in FC frame       */
    #                         /* __u8 value : blocksize. 0 = off      */
    #
    # __u8  stmin;            /* separation time provided in FC frame */
    #                         /* __u8 value :                         */
    #                         /* 0x00 - 0x7F : 0 - 127 ms             */
    #                         /* 0x80 - 0xF0 : reserved               */
    #                         /* 0xF1 - 0xF9 : 100 us - 900 us        */
    #                         /* 0xFA - 0xFF : reserved               */
    #
    # __u8  wftmax;           /* max. number of wait frame transmiss. */
    #                         /* __u8 value : 0 = omit FC N_PDU WT    */
    # };

    def __build_can_isotp_ll_options(self,
                                     mtu=CAN_ISOTP_DEFAULT_LL_MTU,
                                     tx_dl=CAN_ISOTP_DEFAULT_LL_TX_DL,
                                     tx_flags=CAN_ISOTP_DEFAULT_LL_TX_FLAGS
                                     ):
        # type: (int, int, int) -> bytes
        return struct.pack(self.can_isotp_ll_options_fmt,
                           mtu,
                           tx_dl,
                           tx_flags)

    # == Must use native not standard types for packing ==
    # struct can_isotp_ll_options {
    #
    # __u8  mtu;              /* generated & accepted CAN frame type  */
    #                         /* __u8 value :                         */
    #                         /* CAN_MTU   (16) -> standard CAN 2.0   */
    #                         /* CANFD_MTU (72) -> CAN FD frame       */
    #
    # __u8  tx_dl;            /* tx link layer data length in bytes   */
    #                         /* (configured maximum payload length)  */
    #                         /* __u8 value : 8,12,16,20,24,32,48,64  */
    #                         /* => rx path supports all LL_DL values */
    #
    # __u8  tx_flags;         /* set into struct canfd_frame.flags    */
    #                         /* at frame creation: e.g. CANFD_BRS    */
    #                         /* Obsolete when the BRS flag is fixed  */
    #                         /* by the CAN netdriver configuration   */
    # };

    def __get_sock_ifreq(self, sock, iface):
        # type: (socket.socket, str) -> ifreq
        socket_id = ctypes.c_int(sock.fileno())
        ifr = ifreq()
        ifr.ifr_name = iface.encode('ascii')
        ret = LIBC.ioctl(socket_id, SIOCGIFINDEX, ctypes.byref(ifr))

        if ret < 0:
            m = u'Failure while getting "{}" interface index.'.format(
                iface)
            raise Scapy_Exception(m)
        return ifr

    def __bind_socket(self, sock, iface, tx_id, rx_id):
        # type: (socket.socket, str, int, int) -> None
        socket_id = ctypes.c_int(sock.fileno())
        ifr = self.__get_sock_ifreq(sock, iface)

        if tx_id > 0x7ff:
            tx_id = tx_id | socket.CAN_EFF_FLAG
        if rx_id > 0x7ff:
            rx_id = rx_id | socket.CAN_EFF_FLAG

        # select the CAN interface and bind the socket to it
        addr = sockaddr_can(ctypes.c_uint16(socket.PF_CAN),
                            ifr.ifr_ifindex,
                            addr_info(tp(ctypes.c_uint32(rx_id),
                                         ctypes.c_uint32(tx_id))))

        error = LIBC.bind(socket_id, ctypes.byref(addr),
                          ctypes.sizeof(addr))

        if error < 0:
            log_isotp.warning("Couldn't bind socket")

    def __set_option_flags(self,
                           sock,  # type: socket.socket
                           extended_addr=None,  # type: Optional[int]
                           extended_rx_addr=None,  # type: Optional[int]
                           listen_only=False,  # type: bool
                           padding=False,  # type: bool
                           transmit_time=100  # type: int
                           ):
        # type: (...) -> None
        option_flags = CAN_ISOTP_DEFAULT_FLAGS
        if extended_addr is not None:
            option_flags = option_flags | CAN_ISOTP_EXTEND_ADDR
        else:
            extended_addr = CAN_ISOTP_DEFAULT_EXT_ADDRESS

        if extended_rx_addr is not None:
            option_flags = option_flags | CAN_ISOTP_RX_EXT_ADDR
        else:
            extended_rx_addr = CAN_ISOTP_DEFAULT_EXT_ADDRESS

        if listen_only:
            option_flags = option_flags | CAN_ISOTP_LISTEN_MODE

        if padding:
            option_flags = option_flags | CAN_ISOTP_TX_PADDING | CAN_ISOTP_RX_PADDING  # noqa: E501

        sock.setsockopt(SOL_CAN_ISOTP,
                        CAN_ISOTP_OPTS,
                        self.__build_can_isotp_options(
                            frame_txtime=transmit_time,
                            flags=option_flags,
                            ext_address=extended_addr,
                            rx_ext_address=extended_rx_addr))

    def __init__(self,
                 iface=None,  # type: Optional[Union[str, SuperSocket]]
                 tx_id=0,  # type: int
                 rx_id=0,  # type: int
                 ext_address=None,  # type: Optional[int]
                 rx_ext_address=None,  # type: Optional[int]
                 bs=CAN_ISOTP_DEFAULT_RECV_BS,  # type: int
                 stmin=CAN_ISOTP_DEFAULT_RECV_STMIN,  # type: int
                 padding=False,  # type: bool
                 listen_only=False,  # type: bool
                 frame_txtime=CAN_ISOTP_DEFAULT_FRAME_TXTIME,  # type: int
                 fd=False,  # type: bool
                 brs=False,  # type: bool
                 basecls=ISOTP  # type: Type[Packet]
                 ):
        # type: (...) -> None

        if not isinstance(iface, str):
            # This is for interoperability with ISOTPSoftSockets.
            # If a NativeCANSocket is provided, the interface name of this
            # socket is extracted and an ISOTPNativeSocket will be opened
            # on this interface.
            iface = cast(SuperSocket, iface)
            if hasattr(iface, "ins") and hasattr(iface.ins, "getsockname"):
                iface = iface.ins.getsockname()
                if isinstance(iface, tuple):
                    iface = cast(str, iface[0])
            else:
                raise Scapy_Exception("Provide a string or a CANSocket "
                                      "object as iface parameter")

        self.iface: str = cast(str, iface) or conf.contribs['NativeCANSocket']['iface']  # noqa: E501
        # store arguments internally
        self.tx_id = tx_id
        self.rx_id = rx_id
        self.ext_address = ext_address
        self.rx_ext_address = rx_ext_address
        self.bs = bs
        self.stmin = stmin
        self.padding = padding
        self.listen_only = listen_only
        self.frame_txtime = frame_txtime
        self.fd = fd
        self.brs = brs
        if basecls is None:
            log_isotp.warning('Provide a basecls ')
        self.basecls = basecls
        self._init_socket()

    def _init_socket(self) -> None:
        can_socket = socket.socket(
            socket.PF_CAN, socket.SOCK_DGRAM, CAN_ISOTP)

        self.__set_option_flags(
            can_socket,
            self.ext_address,
            self.rx_ext_address,
            self.listen_only,
            self.padding,
            self.frame_txtime)

        can_socket.setsockopt(
            SOL_CAN_ISOTP,
            CAN_ISOTP_RECV_FC,
            self.__build_can_isotp_fc_options(stmin=self.stmin, bs=self.bs))

        tx_flags = ((CANFD_FDF if self.fd else 0) +
                    (CANFD_BRS if (self.brs + self.fd) else 0))
        tx_dl = CAN_FD_ISOTP_DEFAULT_LL_TX_DL if self.fd else CAN_ISOTP_DEFAULT_LL_TX_DL

        can_socket.setsockopt(
            SOL_CAN_ISOTP,
            CAN_ISOTP_LL_OPTS,
            self.__build_can_isotp_ll_options(
                mtu=CAN_ISOTP_CANFD_MTU if self.fd else CAN_ISOTP_DEFAULT_LL_MTU,
                tx_dl=tx_dl,
                tx_flags=tx_flags))
        can_socket.setsockopt(
            socket.SOL_SOCKET,
            SO_TIMESTAMPNS,
            1
        )

        self.__bind_socket(can_socket, self.iface, self.tx_id, self.rx_id)
        # make sure existing sockets are closed,
        # required in case of a reconnect.
        if getattr(self, "outs", None):
            if getattr(self, "ins", None) != self.outs:
                if self.outs and self.outs.fileno() != -1:
                    self.outs.close()
        if getattr(self, "ins", None):
            if self.ins.fileno() != -1:
                self.ins.close()

        self.ins = can_socket
        self.outs = can_socket
        self.closed = False

    def recv_raw(self, x=0xffff):
        # type: (int) -> Tuple[Optional[Type[Packet]], Optional[bytes], Optional[float]]  # noqa: E501
        """
        Receives a packet, then returns a tuple containing
        (cls, pkt_data, time)
        """
        try:
            pkt, _, ts = self._recv_raw(self.ins, x)
        except BlockingIOError:  # noqa: F821
            log_isotp.warning('Captured no data, socket in non-blocking mode.')
            return None, None, None
        except socket.timeout:
            log_isotp.warning('Captured no data, socket read timed out.')
            return None, None, None
        except OSError as e:
            # something bad happened (e.g. the interface went down)
            log_isotp.warning("Captured no data. %s" % e)
            if e.errno == 84:
                log_isotp.warning("Maybe a consecutive frame was missed. "
                                  "Increasing `stmin` could solve this problem.")
            elif e.errno == 110:
                log_isotp.warning('Captured no data, socket read timed out.')
            elif e.errno == 70:
                log_isotp.warning(
                    'Communication error on send. '
                    'TX path flowcontrol reception timeout.')
            else:
                log_isotp.error(
                    'Unknown error code received %d. Closing socket!', e.errno)
                self.close()
            return None, None, None

        if pkt and ts is None:
            ts = get_last_packet_timestamp(self.ins)
        return self.basecls, pkt, ts

    def recv(self, x=0xffff, **kwargs):
        # type: (int, **Any) -> Optional[Packet]
        msg = SuperSocket.recv(self, x, **kwargs)
        if msg is None:
            return msg

        if hasattr(msg, "tx_id"):
            msg.tx_id = self.tx_id
        if hasattr(msg, "rx_id"):
            msg.rx_id = self.rx_id
        if hasattr(msg, "ext_address"):
            msg.ext_address = self.ext_address
        if hasattr(msg, "rx_ext_address"):
            msg.rx_ext_address = self.rx_ext_address
        return msg



# SPDX-License-Identifier: GPL-2.0-only
# This file is part of Scapy
# See https://scapy.net/ for more information
# Copyright (C) Gabriel Potter

"""
[MS-PAC]

https://docs.microsoft.com/en-us/openspecs/windows_protocols/ms-pac/166d8064-c863-41e1-9c23-edaaa5f36962
Up to date with version: 23.0
"""

import struct

from scapy.config import conf
from scapy.error import log_runtime
from scapy.fields import (
    ConditionalField,
    FieldLenField,
    FieldListField,
    FlagsField,
    LEIntEnumField,
    LEIntField,
    LELongField,
    LEShortField,
    MultipleTypeField,
    PacketField,
    PacketListField,
    StrField,
    StrFieldUtf16,
    StrFixedLenField,
    StrLenFieldUtf16,
    UTCTimeField,
    UUIDField,
    XStrField,
    XStrLenField,
)
from scapy.packet import Packet
from scapy.layers.kerberos import (
    _AUTHORIZATIONDATA_VALUES,
    _KRB_S_TYPES,
)
from scapy.layers.dcerpc import (
    NDRByteField,
    NDRConfFieldListField,
    NDRConfPacketListField,
    NDRConfStrLenField,
    NDRConfVarStrLenFieldUtf16,
    NDRConfVarStrNullFieldUtf16,
    NDRConformantString,
    NDRFieldListField,
    NDRFullEmbPointerField,
    NDRInt3264EnumField,
    NDRIntField,
    NDRLongField,
    NDRPacket,
    NDRPacketField,
    NDRSerialization1Header,
    NDRSerializeType1PacketLenField,
    NDRShortField,
    NDRSignedLongField,
    NDRUnionField,
    _NDRConfField,
    ndr_deserialize1,
    ndr_serialize1,
)
from scapy.layers.ntlm import (
    _NTLMPayloadField,
    _NTLMPayloadPacket,
)
from scapy.layers.smb2 import WINNT_SID

# sect 2.4


class PAC_INFO_BUFFER(Packet):
    fields_desc = [
        LEIntEnumField(
            "ulType",
            0x00000001,
            {
                0x00000001: "Logon information",
                0x00000002: "Credentials information",
                0x00000006: "Server Signature",
                0x00000007: "KDC Signature",
                0x0000000A: "Client name and ticket information",
                0x0000000B: "Constrained delegation information",
                0x0000000C: "UPN and DNS information",
                0x0000000D: "Client claims information",
                0x0000000E: "Device information",
                0x0000000F: "Device claims information",
                0x00000010: "Ticket Signature",
                0x00000011: "PAC Attributes",
                0x00000012: "PAC Requestor",
                0x00000013: "Extended KDC Signature",
            },
        ),
        LEIntField("cbBufferSize", None),
        LELongField("Offset", None),
    ]

    def default_payload_class(self, payload):
        return conf.padding_layer


_PACTYPES = {}


# sect 2.5 - NDR PACKETS


class RPC_UNICODE_STRING(NDRPacket):
    ALIGNMENT = (4, 8)
    fields_desc = [
        NDRShortField("Length", None, size_of="Buffer", adjust=lambda _, x: (x * 2)),
        NDRShortField(
            "MaximumLength", None, size_of="Buffer", adjust=lambda _, x: (x * 2)
        ),
        NDRFullEmbPointerField(
            NDRConfVarStrLenFieldUtf16(
                "Buffer",
                "",
                size_is=lambda pkt: (pkt.MaximumLength // 2),
                length_is=lambda pkt: (pkt.Length // 2),
            ),
        ),
    ]


class FILETIME(NDRPacket):
    ALIGNMENT = (4, 4)
    fields_desc = [NDRIntField("dwLowDateTime", 0), NDRIntField("dwHighDateTime", 0)]


class GROUP_MEMBERSHIP(NDRPacket):
    ALIGNMENT = (4, 4)
    fields_desc = [NDRIntField("RelativeId", 0), NDRIntField("Attributes", 0)]


class CYPHER_BLOCK(NDRPacket):
    fields_desc = [StrFixedLenField("data", "", length=8)]


class USER_SESSION_KEY(NDRPacket):
    fields_desc = [PacketListField("data", [], CYPHER_BLOCK, count_from=lambda _: 2)]


class RPC_SID_IDENTIFIER_AUTHORITY(NDRPacket):
    fields_desc = [StrFixedLenField("Value", "", length=6)]


class SID(NDRPacket):
    ALIGNMENT = (4, 8)
    DEPORTED_CONFORMANTS = ["SubAuthority"]
    fields_desc = [
        NDRByteField("Revision", 0),
        NDRByteField("SubAuthorityCount", None, size_of="SubAuthority"),
        NDRPacketField(
            "IdentifierAuthority",
            RPC_SID_IDENTIFIER_AUTHORITY(),
            RPC_SID_IDENTIFIER_AUTHORITY,
        ),
        NDRConfFieldListField(
            "SubAuthority",
            [],
            NDRIntField("", 0),
            size_is=lambda pkt: pkt.SubAuthorityCount,
            conformant_in_struct=True,
        ),
    ]

    def summary(self):
        return WINNT_SID.summary(self)


class KERB_SID_AND_ATTRIBUTES(NDRPacket):
    ALIGNMENT = (4, 8)
    fields_desc = [
        NDRFullEmbPointerField(NDRPacketField("Sid", SID(), SID)),
        NDRIntField("Attributes", 0),
    ]


class KERB_VALIDATION_INFO(NDRPacket):
    ALIGNMENT = (4, 8)
    fields_desc = [
        NDRPacketField("LogonTime", FILETIME(), FILETIME),
        NDRPacketField("LogoffTime", FILETIME(), FILETIME),
        NDRPacketField("KickOffTime", FILETIME(), FILETIME),
        NDRPacketField("PasswordLastSet", FILETIME(), FILETIME),
        NDRPacketField("PasswordCanChange", FILETIME(), FILETIME),
        NDRPacketField("PasswordMustChange", FILETIME(), FILETIME),
        NDRPacketField("EffectiveName", RPC_UNICODE_STRING(), RPC_UNICODE_STRING),
        NDRPacketField("FullName", RPC_UNICODE_STRING(), RPC_UNICODE_STRING),
        NDRPacketField("LogonScript", RPC_UNICODE_STRING(), RPC_UNICODE_STRING),
        NDRPacketField("ProfilePath", RPC_UNICODE_STRING(), RPC_UNICODE_STRING),
        NDRPacketField("HomeDirectory", RPC_UNICODE_STRING(), RPC_UNICODE_STRING),
        NDRPacketField("HomeDirectoryDrive", RPC_UNICODE_STRING(), RPC_UNICODE_STRING),
        NDRShortField("LogonCount", 0),
        NDRShortField("BadPasswordCount", 0),
        NDRIntField("UserId", 0),
        NDRIntField("PrimaryGroupId", 0),
        NDRIntField("GroupCount", None, size_of="GroupIds"),
        NDRFullEmbPointerField(
            NDRConfPacketListField(
                "GroupIds",
                [GROUP_MEMBERSHIP()],
                GROUP_MEMBERSHIP,
                size_is=lambda pkt: pkt.GroupCount,
            ),
        ),
        NDRIntField("UserFlags", 0),
        NDRPacketField("UserSessionKey", USER_SESSION_KEY(), USER_SESSION_KEY),
        NDRPacketField("LogonServer", RPC_UNICODE_STRING(), RPC_UNICODE_STRING),
        NDRPacketField("LogonDomainName", RPC_UNICODE_STRING(), RPC_UNICODE_STRING),
        NDRFullEmbPointerField(NDRPacketField("LogonDomainId", SID(), SID)),
        NDRFieldListField("Reserved1", [], NDRIntField("", 0), length_is=lambda _: 2),
        NDRIntField("UserAccountControl", 0),
        NDRFieldListField("Reserved3", [], NDRIntField("", 0), length_is=lambda _: 7),
        NDRIntField("SidCount", None, size_of="ExtraSids"),
        NDRFullEmbPointerField(
            NDRConfPacketListField(
                "ExtraSids",
                [KERB_SID_AND_ATTRIBUTES()],
                KERB_SID_AND_ATTRIBUTES,
                size_is=lambda pkt: pkt.SidCount,
            ),
        ),
        NDRFullEmbPointerField(
            NDRPacketField("ResourceGroupDomainSid", SID(), SID),
        ),
        NDRIntField("ResourceGroupCount", None, size_of="ResourceGroupIds"),
        NDRFullEmbPointerField(
            NDRConfPacketListField(
                "ResourceGroupIds",
                [GROUP_MEMBERSHIP()],
                GROUP_MEMBERSHIP,
                size_is=lambda pkt: pkt.ResourceGroupCount,
            ),
        ),
    ]


_PACTYPES[1] = KERB_VALIDATION_INFO

# sect 2.6


class PAC_CREDENTIAL_INFO(Packet):
    fields_desc = [
        LEIntField("Version", 0),
        LEIntEnumField(
            "EncryptionType",
            1,
            {
                0x00000001: "DES-CBC-CRC",
                0x00000003: "DES-CBC-MD5",
                0x00000011: "AES128_CTS_HMAC_SHA1_96",
                0x00000012: "AES256_CTS_HMAC_SHA1_96",
                0x00000017: "RC4-HMAC",
            },
        ),
        XStrField("SerializedData", b""),
    ]


_PACTYPES[2] = PAC_CREDENTIAL_INFO

# sect 2.7


class PAC_CLIENT_INFO(Packet):
    fields_desc = [
        UTCTimeField(
            "ClientId", None, fmt="<Q", epoch=[1601, 1, 1, 0, 0, 0], custom_scaling=1e7
        ),
        FieldLenField("NameLength", None, length_of="Name", fmt="<H"),
        StrLenFieldUtf16("Name", b"", length_from=lambda pkt: pkt.NameLength),
    ]


_PACTYPES[0xA] = PAC_CLIENT_INFO

# sect 2.8


class PAC_SIGNATURE_DATA(Packet):
    fields_desc = [
        LEIntEnumField(
            "SignatureType",
            None,
            _KRB_S_TYPES,
        ),
        XStrLenField(
            "Signature",
            b"",
            length_from=lambda pkt: {
                0x1: 4,
                0xFFFFFF76: 16,
                0x0000000F: 12,
                0x00000010: 12,
            }.get(pkt.SignatureType, 0),
        ),
        StrField("RODCIdentifier", b""),
    ]


_PACTYPES[6] = PAC_SIGNATURE_DATA
_PACTYPES[7] = PAC_SIGNATURE_DATA
_PACTYPES[0x10] = PAC_SIGNATURE_DATA
_PACTYPES[0x13] = PAC_SIGNATURE_DATA

# sect 2.9


class S4U_DELEGATION_INFO(NDRPacket):
    ALIGNMENT = (4, 8)
    fields_desc = [
        NDRPacketField("S4U2proxyTarget", RPC_UNICODE_STRING(), RPC_UNICODE_STRING),
        NDRIntField("TransitedListSize", None, size_of="S4UTransitedServices"),
        NDRFullEmbPointerField(
            NDRConfPacketListField(
                "S4UTransitedServices",
                [RPC_UNICODE_STRING()],
                RPC_UNICODE_STRING,
                size_is=lambda pkt: pkt.TransitedListSize,
            ),
        ),
    ]


# sect 2.10


def _pac_post_build(self, p, pay_offset, fields):
    """Util function to build the offset and populate the lengths"""
    for field_name, value in self.fields["Payload"]:
        length = self.get_field("Payload").fields_map[field_name].i2len(self, value)
        offset = fields[field_name]
        # Length
        if self.getfieldval(field_name + "Len") is None:
            p = p[:offset] + struct.pack("<H", length) + p[offset + 2 :]
        # Offset
        if self.getfieldval(field_name + "BufferOffset") is None:
            p = p[: offset + 2] + struct.pack("<H", pay_offset) + p[offset + 4 :]
        pay_offset += length
    return p


class UPN_DNS_INFO(_NTLMPayloadPacket):
    fields_desc = [
        LEShortField("UpnLen", None),
        LEShortField("UpnBufferOffset", None),
        LEShortField("DnsDomainNameLen", None),
        LEShortField("DnsDomainNameBufferOffset", None),
        FlagsField(
            "Flags",
            0,
            -32,
            [
                "U",
                "S",  # Extended
            ],
        ),
        ConditionalField(
            # Extended
            LEShortField("SamNameLen", None),
            lambda pkt: pkt.Flags.S,
        ),
        ConditionalField(
            # Extended
            LEShortField("SamNameBufferOffset", None),
            lambda pkt: pkt.Flags.S,
        ),
        ConditionalField(
            # Extended
            LEShortField("SidLen", None),
            lambda pkt: pkt.Flags.S,
        ),
        ConditionalField(
            # Extended
            LEShortField("SidBufferOffset", None),
            lambda pkt: pkt.Flags.S,
        ),
        MultipleTypeField(
            [
                (
                    # Extended
                    _NTLMPayloadField(
                        "Payload",
                        20,
                        [
                            StrFieldUtf16("Upn", b""),
                            StrFieldUtf16("DnsDomainName", b""),
                            StrFieldUtf16("SamName", b""),
                            PacketField("Sid", WINNT_SID(), WINNT_SID),
                        ],
                    ),
                    lambda pkt: pkt.Flags.S,
                )
            ],
            # Not-extended
            _NTLMPayloadField(
                "Payload",
                12,
                [
                    StrFieldUtf16("Upn", b""),
                    StrFieldUtf16("DnsDomainName", b""),
                ],
            ),
        ),
    ]

    def post_build(self, pkt, pay):
        # type: (bytes, bytes) -> bytes
        offset = 12
        fields = {
            "Upn": 0,
            "DnsDomainName": 4,
        }
        if self.Flags.S:
            offset = 20
            fields["SamName"] = 12
            fields["Sid"] = 16
        return (
            _pac_post_build(
                self,
                pkt,
                offset,
                fields,
            )
            + pay
        )


_PACTYPES[0xC] = UPN_DNS_INFO

# sect 2.11 - NDR PACKETS

try:
    from enum import IntEnum
except ImportError:
    IntEnum = object


class CLAIM_TYPE(IntEnum):
    CLAIM_TYPE_INT64 = 1
    CLAIM_TYPE_UINT64 = 2
    CLAIM_TYPE_STRING = 3
    CLAIM_TYPE_BOOLEAN = 6


class CLAIMS_SOURCE_TYPE(IntEnum):
    CLAIMS_SOURCE_TYPE_AD = 1
    CLAIMS_SOURCE_TYPE_CERTIFICATE = 2


class CLAIMS_COMPRESSION_FORMAT(IntEnum):
    COMPRESSION_FORMAT_NONE = 0
    COMPRESSION_FORMAT_LZNT1 = 2
    COMPRESSION_FORMAT_XPRESS = 3
    COMPRESSION_FORMAT_XPRESS_HUFF = 4


class CLAIM_ENTRY_sub0(NDRPacket):
    ALIGNMENT = (4, 8)
    fields_desc = [
        NDRIntField("ValueCount", None, size_of="Int64Values"),
        NDRFullEmbPointerField(
            NDRConfFieldListField(
                "Int64Values",
                [],
                NDRSignedLongField,
                size_is=lambda pkt: pkt.ValueCount,
            ),
        ),
    ]


class CLAIM_ENTRY_sub1(NDRPacket):
    ALIGNMENT = (4, 8)
    fields_desc = [
        NDRIntField("ValueCount", None, size_of="Uint64Values"),
        NDRFullEmbPointerField(
            NDRConfFieldListField(
                "Uint64Values", [], NDRLongField, size_is=lambda pkt: pkt.ValueCount
            ),
        ),
    ]


class CLAIM_ENTRY_sub2(NDRPacket):
    ALIGNMENT = (4, 8)
    fields_desc = [
        NDRIntField("ValueCount", None, size_of="StringValues"),
        NDRFullEmbPointerField(
            NDRConfFieldListField(
                "StringValues",
                [],
                NDRFullEmbPointerField(
                    NDRConfVarStrNullFieldUtf16("StringVal", ""),
                ),
                size_is=lambda pkt: pkt.ValueCount,
            ),
        ),
    ]


class CLAIM_ENTRY_sub3(NDRPacket):
    ALIGNMENT = (4, 8)
    fields_desc = [
        NDRIntField("ValueCount", None, size_of="BooleanValues"),
        NDRFullEmbPointerField(
            NDRConfFieldListField(
                "BooleanValues", [], NDRLongField, size_is=lambda pkt: pkt.ValueCount
            ),
        ),
    ]


class CLAIM_ENTRY(NDRPacket):
    ALIGNMENT = (4, 8)
    fields_desc = [
        NDRFullEmbPointerField(NDRConfVarStrNullFieldUtf16("Id", "")),
        NDRInt3264EnumField("Type", 0, CLAIM_TYPE),
        NDRUnionField(
            [
                (
                    NDRPacketField("Values", CLAIM_ENTRY_sub0(), CLAIM_ENTRY_sub0),
                    (
                        (
                            lambda pkt: getattr(pkt, "Type", None)
                            == CLAIM_TYPE.CLAIM_TYPE_INT64
                        ),
                        (lambda _, val: val.tag == CLAIM_TYPE.CLAIM_TYPE_INT64),
                    ),
                ),
                (
                    NDRPacketField("Values", CLAIM_ENTRY_sub1(), CLAIM_ENTRY_sub1),
                    (
                        (
                            lambda pkt: getattr(pkt, "Type", None)
                            == CLAIM_TYPE.CLAIM_TYPE_UINT64
                        ),
                        (lambda _, val: val.tag == CLAIM_TYPE.CLAIM_TYPE_UINT64),
                    ),
                ),
                (
                    NDRPacketField("Values", CLAIM_ENTRY_sub2(), CLAIM_ENTRY_sub2),
                    (
                        (
                            lambda pkt: getattr(pkt, "Type", None)
                            == CLAIM_TYPE.CLAIM_TYPE_STRING
                        ),
                        (lambda _, val: val.tag == CLAIM_TYPE.CLAIM_TYPE_STRING),
                    ),
                ),
                (
                    NDRPacketField("Values", CLAIM_ENTRY_sub3(), CLAIM_ENTRY_sub3),
                    (
                        (
                            lambda pkt: getattr(pkt, "Type", None)
                            == CLAIM_TYPE.CLAIM_TYPE_BOOLEAN
                        ),
                        (lambda _, val: val.tag == CLAIM_TYPE.CLAIM_TYPE_BOOLEAN),
                    ),
                ),
            ],
            StrFixedLenField("Values", "", length=0),
            align=(2, 8),
            switch_fmt=("H", "I"),
        ),
    ]


class CLAIMS_ARRAY(NDRPacket):
    ALIGNMENT = (4, 8)
    fields_desc = [
        NDRInt3264EnumField("usClaimsSourceType", 0, CLAIMS_SOURCE_TYPE),
        NDRIntField("ulClaimsCount", None, size_of="ClaimEntries"),
        NDRFullEmbPointerField(
            NDRConfPacketListField(
                "ClaimEntries",
                [CLAIM_ENTRY()],
                CLAIM_ENTRY,
                size_is=lambda pkt: pkt.ulClaimsCount,
            ),
        ),
    ]


class CLAIMS_SET(NDRPacket):
    ALIGNMENT = (4, 8)
    fields_desc = [
        NDRIntField("ulClaimsArrayCount", None, size_of="ClaimsArrays"),
        NDRFullEmbPointerField(
            NDRConfPacketListField(
                "ClaimsArrays",
                [CLAIMS_ARRAY()],
                CLAIMS_ARRAY,
                size_is=lambda pkt: pkt.ulClaimsArrayCount,
            ),
        ),
        NDRShortField("usReservedType", 0),
        NDRIntField("ulReservedFieldSize", None, size_of="ReservedField"),
        NDRFullEmbPointerField(
            NDRConfStrLenField(
                "ReservedField", "", size_is=lambda pkt: pkt.ulReservedFieldSize
            ),
        ),
    ]


class _CLAIMSClaimSet(_NDRConfField, NDRSerializeType1PacketLenField):
    CONFORMANT_STRING = True
    LENGTH_FROM = True

    def m2i(self, pkt, s):
        if pkt.usCompressionFormat == CLAIMS_COMPRESSION_FORMAT.COMPRESSION_FORMAT_NONE:
            return ndr_deserialize1(s, CLAIMS_SET, ptr_pack=True)
        else:
            # TODO: There are 3 funky compression formats... see sect 2.2.18.4
            return NDRConformantString(value=s)

    def i2m(self, pkt, val):
        val = val[0]
        if pkt.usCompressionFormat == CLAIMS_COMPRESSION_FORMAT.COMPRESSION_FORMAT_NONE:
            return ndr_serialize1(val, ptr_pack=True)
        else:
            # funky
            return bytes(val)

    def valueof(self, pkt, x):
        if pkt.usCompressionFormat == CLAIMS_COMPRESSION_FORMAT.COMPRESSION_FORMAT_NONE:
            return self._subval(x)[0]
        else:
            return x


class CLAIMS_SET_METADATA(NDRPacket):
    ALIGNMENT = (4, 8)
    fields_desc = [
        NDRIntField("ulClaimsSetSize", None, size_of="ClaimsSet"),
        NDRFullEmbPointerField(
            _CLAIMSClaimSet(
                "ClaimsSet", None, None, size_is=lambda pkt: pkt.ulClaimsSetSize
            ),
        ),
        NDRInt3264EnumField(
            "usCompressionFormat",
            0,
            CLAIMS_COMPRESSION_FORMAT,
        ),
        # this size_of is technically wrong. we just assume it's uncompressed...
        NDRIntField("ulUncompressedClaimsSetSize", None, size_of="ClaimsSet"),
        NDRShortField("usReservedType", 0),
        NDRIntField("ulReservedFieldSize", None, size_of="ReservedField"),
        NDRFullEmbPointerField(
            NDRConfStrLenField(
                "ReservedField", "", size_is=lambda pkt: pkt.ulReservedFieldSize
            ),
        ),
    ]


class PAC_CLIENT_CLAIMS_INFO(NDRPacket):
    fields_desc = [NDRPacketField("Claims", CLAIMS_SET_METADATA(), CLAIMS_SET_METADATA)]


if IntEnum != object:
    # If not available, ignore. I can't be bothered
    _PACTYPES[0xD] = PAC_CLIENT_CLAIMS_INFO


# sect 2.12 - NDR PACKETS


class DOMAIN_GROUP_MEMBERSHIP(NDRPacket):
    ALIGNMENT = (4, 8)
    fields_desc = [
        NDRFullEmbPointerField(NDRPacketField("DomainId", SID(), SID)),
        NDRIntField("GroupCount", 0),
        NDRFullEmbPointerField(
            NDRConfPacketListField(
                "GroupIds",
                [GROUP_MEMBERSHIP()],
                GROUP_MEMBERSHIP,
                size_is=lambda pkt: pkt.GroupCount,
            ),
        ),
    ]


class PAC_DEVICE_INFO(NDRPacket):
    ALIGNMENT = (4, 8)
    fields_desc = [
        NDRIntField("UserId", 0),
        NDRIntField("PrimaryGroupId", 0),
        NDRFullEmbPointerField(
            NDRPacketField("AccountDomainId", SID(), SID),
        ),
        NDRIntField("AccountGroupCount", 0),
        NDRFullEmbPointerField(
            NDRConfPacketListField(
                "AccountGroupIds",
                [GROUP_MEMBERSHIP()],
                GROUP_MEMBERSHIP,
                size_is=lambda pkt: pkt.AccountGroupCount,
            ),
        ),
        NDRIntField("SidCount", 0),
        NDRFullEmbPointerField(
            NDRConfPacketListField(
                "ExtraSids",
                [KERB_SID_AND_ATTRIBUTES()],
                KERB_SID_AND_ATTRIBUTES,
                size_is=lambda pkt: pkt.SidCount,
            ),
        ),
        NDRIntField("DomainGroupCount", 0),
        NDRFullEmbPointerField(
            NDRConfPacketListField(
                "DomainGroup",
                [DOMAIN_GROUP_MEMBERSHIP()],
                DOMAIN_GROUP_MEMBERSHIP,
                size_is=lambda pkt: pkt.DomainGroupCount,
            ),
        ),
    ]


_PACTYPES[0xE] = PAC_DEVICE_INFO

# sect 2.14 - PAC_ATTRIBUTES_INFO


class PAC_ATTRIBUTES_INFO(Packet):
    fields_desc = [
        LEIntField("FlagsLength", 2),
        FieldListField(
            "Flags",
            ["PAC_WAS_REQUESTED"],
            FlagsField(
                "",
                0,
                -32,
                {
                    0x00000001: "PAC_WAS_REQUESTED",
                    0x00000002: "PAC_WAS_GIVEN_IMPLICITLY",
                },
            ),
            count_from=lambda pkt: (pkt.FlagsLength + 7) // 8,
        ),
    ]


_PACTYPES[0x11] = PAC_ATTRIBUTES_INFO

# sect 2.15 - PAC_REQUESTOR_SID


class PAC_REQUESTOR_SID(Packet):
    fields_desc = [
        PacketField("Sid", WINNT_SID(), WINNT_SID),
    ]


_PACTYPES[0x12] = PAC_REQUESTOR_SID

# sect 2.16 - PAC_REQUESTOR_GUID


class PAC_REQUESTOR_GUID(Packet):
    fields_desc = [
        UUIDField("Guid", None),
    ]


_PACTYPES[0x14] = PAC_REQUESTOR_GUID

# sect 2.3


class _PACTYPEBuffers(PacketListField):
    def addfield(self, pkt, s, val):
        # we use this field to set Offset and cbBufferSize
        res = b""
        if len(val) != len(pkt.Payloads):
            log_runtime.warning("Size of 'Buffers' does not match size of 'Payloads' !")
            return super(_PACTYPEBuffers, self).addfield(pkt, s, val)
        offset = 16 * len(pkt.Payloads) + 8
        for i, v in enumerate(val):
            x = self.i2m(pkt, v)
            pay = pkt.Payloads[i]
            if isinstance(pay, NDRPacket) or isinstance(pay, NDRSerialization1Header):
                lgth = len(ndr_serialize1(pay, ptr_pack=True))
            else:
                lgth = len(pay)
            if v.cbBufferSize is None:
                x = x[:4] + struct.pack("<I", lgth) + x[8:]
            if v.Offset is None:
                x = x[:8] + struct.pack("<Q", offset) + x[16:]
            offset += lgth
            offset += (-offset) % 8  # Account for padding
            res += x
        return s + res


class _PACTYPEPayloads(PacketListField):
    def i2m(self, pkt, val):
        if isinstance(val, NDRPacket) or isinstance(val, NDRSerialization1Header):
            s = ndr_serialize1(val, ptr_pack=True)
        else:
            s = bytes(val)
        return s + b"\x00" * ((-len(s)) % 8)

    def getfield(self, pkt, s):
        if not pkt or not s:
            return s, []
        result = []
        for buf in pkt.Buffers:
            offset = buf.Offset - 16 * len(pkt.Buffers) - 8
            try:
                cls = _PACTYPES[buf.ulType]
                if buf.cbBufferSize == 0:
                    # empty size
                    raise KeyError
                if issubclass(cls, NDRPacket):
                    val = ndr_deserialize1(
                        s[offset : offset + buf.cbBufferSize],
                        cls,
                        ptr_pack=True,
                    )
                else:
                    val = cls(s[offset : offset + buf.cbBufferSize])
                if conf.raw_layer in val:
                    pad = conf.padding_layer(load=val[conf.raw_layer].load)
                    lay = val[conf.raw_layer].underlayer
                    if not lay:
                        val.show()
                        raise ValueError("Dissection failed")
                    lay.remove_payload()
                    lay.add_payload(pad)
            except KeyError:
                val = conf.padding_layer(s[offset : offset + buf.cbBufferSize])
            result.append(val)
        return b"", result


class PACTYPE(Packet):
    name = "PACTYPE - PAC"
    fields_desc = [
        FieldLenField("cBuffers", None, count_of="Buffers", fmt="<I"),
        LEIntField("Version", 0x00000000),
        _PACTYPEBuffers(
            "Buffers",
            [PAC_INFO_BUFFER()],
            PAC_INFO_BUFFER,
            count_from=lambda pkt: pkt.cBuffers,
        ),
        _PACTYPEPayloads("Payloads", [], None),
    ]

    def getPayload(self, ulType):
        """
        Get a payload if it exists.
        """
        for i, buf in enumerate(self.Buffers):
            if buf.ulType == ulType:
                return self.Payloads[i]
        return None


_AUTHORIZATIONDATA_VALUES[128] = PACTYPE  # AD-WIN2K-PAC



# SPDX-License-Identifier: GPL-2.0-or-later
# This file is part of Scapy
# See https://scapy.net/ for more information
# Copyright (C) Gabriel Potter <gabriel[]potter[]fr>

"""
Generate the bluetoothids.py file based on blueooth_sig's public listing
"""

import yaml
import json
import gzip
import urllib.request

from base64 import b85encode

URL = "https://bitbucket.org/bluetooth-SIG/public/raw/main/assigned_numbers/company_identifiers/company_identifiers.yaml"  # noqa: E501

with urllib.request.urlopen(URL) as stream:
    DATA = yaml.safe_load(stream.read())

COMPILED = {}

for company in DATA["company_identifiers"]:
    COMPILED[company["value"]] = company["name"]

# Compress properly
COMPILED = gzip.compress(json.dumps(COMPILED).encode())
# Encode in Base85
COMPILED = b85encode(COMPILED).decode()
# Split
COMPILED = "\n".join(COMPILED[i : i + 79] for i in range(0, len(COMPILED), 79)) + "\n"


with open("../libs/bluetoothids.py", "r") as inp:
    data = inp.read()

with open("../libs/bluetoothids.py", "w") as out:
    ini, sep, _ = data.partition("DATA = _d(\"\"\"")
    COMPILED = ini + sep + "\n" + COMPILED + "\"\"\")\n"
    print("Written: %s" % out.write(COMPILED))



# SPDX-License-Identifier: GPL-2.0-only
# This file is part of Scapy
# See https://scapy.net/ for more information
# Copyright (C) Nils Weiss <nils@we155.de>

# scapy.contrib.description = GMLAN AutomotiveTestCaseExecutor Utilities
# scapy.contrib.status = loads

import abc
import random
import time
import copy

from collections import defaultdict

from scapy.compat import orb
from scapy.contrib.automotive import log_automotive
from scapy.packet import Packet
from scapy.config import conf
from scapy.supersocket import SuperSocket
from scapy.error import Scapy_Exception
from scapy.contrib.automotive.gm.gmlanutils import GMLAN_InitDiagnostics, \
    GMLAN_TesterPresentSender
from scapy.contrib.automotive.gm.gmlan import GMLAN, GMLAN_SA, GMLAN_RD, \
    GMLAN_TD, GMLAN_RMBA, GMLAN_RDBI, GMLAN_RDBPI, GMLAN_IDO, \
    GMLAN_NR, GMLAN_WDBI, GMLAN_DC, GMLAN_PM
from scapy.contrib.automotive.ecu import EcuState

from scapy.contrib.automotive.scanner.test_case import AutomotiveTestCaseABC, \
    _SocketUnion, _TransitionTuple, StateGenerator
from scapy.contrib.automotive.scanner.enumerator import ServiceEnumerator, \
    _AutomotiveTestCaseScanResult, StateGeneratingServiceEnumerator
from scapy.contrib.automotive.scanner.configuration import \
    AutomotiveTestCaseExecutorConfiguration
from scapy.contrib.automotive.scanner.graph import _Edge
from scapy.contrib.automotive.scanner.staged_test_case import \
    StagedAutomotiveTestCase
from scapy.contrib.automotive.scanner.executor import \
    AutomotiveTestCaseExecutor

# TODO: Refactor this import
from scapy.contrib.automotive.gm.gmlan_ecu_states import *  # noqa: F401, F403

# Typing imports
from typing import (
    Optional,
    List,
    Type,
    Any,
    Tuple,
    Iterable,
    Dict,
    cast,
    Callable,
)

__all__ = ["GMLAN_Scanner", "GMLAN_ServiceEnumerator", "GMLAN_RDBIEnumerator",
           "GMLAN_RDBPIEnumerator", "GMLAN_RMBAEnumerator",
           "GMLAN_TPEnumerator", "GMLAN_IDOEnumerator", "GMLAN_PMEnumerator",
           "GMLAN_RDEnumerator", "GMLAN_TDEnumerator", "GMLAN_WDBIEnumerator",
           "GMLAN_SAEnumerator", "GMLAN_WDBISelectiveEnumerator",
           "GMLAN_DCEnumerator"]


class GMLAN_Enumerator(ServiceEnumerator, metaclass=abc.ABCMeta):
    """
    Abstract base class for GMLAN service enumerators. This class
    implements GMLAN specific functions.
    """
    @staticmethod
    def _get_negative_response_code(resp):
        # type: (Packet) -> int
        return resp.returnCode

    @staticmethod
    def _get_negative_response_desc(nrc):
        # type: (int) -> str
        return GMLAN_NR(returnCode=nrc).sprintf("%GMLAN_NR.returnCode%")

    @staticmethod
    def _get_negative_response_label(response):
        # type: (Packet) -> str
        return response.sprintf("NR: %GMLAN_NR.returnCode%")

    def _get_table_entry_z(self, tup):
        # type: (_AutomotiveTestCaseScanResult) -> str
        return self._get_label(tup[2])

    def _get_initial_requests(self, **kwargs):
        # type: (Any) -> Iterable[Packet]
        raise NotImplementedError("Overwrite this method")


class GMLAN_ServiceEnumerator(GMLAN_Enumerator, StateGeneratingServiceEnumerator):  # noqa: E501
    """
    This enumerator scans for all services identifiers of GMLAN. During this
    scan, corrupted packets might be sent to an ECU and mainly negative
    responses will be received.
    """
    _description = "Available services and negative response per state"

    def _get_initial_requests(self, **kwargs):
        # type: (Any) -> Iterable[Packet]
        services = set(x & ~0x40 for x in range(0x100))
        services.remove(0x10)  # Remove InitiateDiagnosticOperation service
        services.remove(0x3E)  # Remove TesterPresent service
        services.remove(0xa5)  # Remove ProgrammingMode service
        services.remove(0x34)  # Remove RequestDownload
        return (GMLAN(service=x) for x in services)

    def _get_table_entry_y(self, tup):
        # type: (_AutomotiveTestCaseScanResult) -> str
        return "0x%02x: %s" % (
            tup[1].service, tup[1].sprintf("%GMLAN.service%"))


class GMLAN_TPEnumerator(GMLAN_Enumerator, StateGeneratingServiceEnumerator):
    """
    Performs a check if TesterPresent is available. If a positive response is
    received, a new system state is generated and returned.
    """
    _description = "TesterPresent supported"

    def _get_initial_requests(self, **kwargs):
        # type: (Any) -> Iterable[Packet]
        return [GMLAN(service=0x3E)]

    @staticmethod
    def enter(socket,  # type: _SocketUnion
              configuration,  # type: AutomotiveTestCaseExecutorConfiguration
              kwargs  # type: Dict[str, Any]
              ):
        # type: (...) -> bool
        if configuration.unittest:
            configuration["tps"] = None
            socket.sr1(GMLAN(service=0x3E), timeout=0.1, verbose=False)
            return True

        GMLAN_TPEnumerator.cleanup(socket, configuration)
        configuration["tps"] = GMLAN_TesterPresentSender(
            cast(SuperSocket, socket))
        configuration["tps"].start()
        return True

    @staticmethod
    def cleanup(_, configuration):
        # type: (_SocketUnion, AutomotiveTestCaseExecutorConfiguration) -> bool
        try:
            if configuration["tps"]:
                configuration["tps"].stop()
                configuration["tps"] = None
        except KeyError:
            pass
        return True

    def get_transition_function(self, socket, edge):
        # type: (_SocketUnion, _Edge) -> Optional[_TransitionTuple]
        return self.enter, {"desc": "TP"}, self.cleanup

    def _get_table_entry_y(self, tup):
        # type: (_AutomotiveTestCaseScanResult) -> str
        return "TesterPresent:"


class GMLAN_IDOEnumerator(GMLAN_Enumerator, StateGeneratingServiceEnumerator):
    _description = "InitiateDiagnosticOperation supported"

    def _get_initial_requests(self, **kwargs):
        # type: (Any) -> Iterable[Packet]
        return [GMLAN() / GMLAN_IDO(subfunction=2)]

    @staticmethod
    def enter_diagnostic_session(socket):
        # type: (_SocketUnion) -> bool
        ans = socket.sr1(
            GMLAN() / GMLAN_IDO(subfunction=2), timeout=5, verbose=False)
        if ans is not None and ans.service == 0x7f:
            log_automotive.debug(
                "InitiateDiagnosticOperation received negative response!\n"
                "%s", repr(ans))
        return ans is not None and ans.service != 0x7f

    def get_new_edge(self, socket, config):
        # type: (_SocketUnion, AutomotiveTestCaseExecutorConfiguration) -> Optional[_Edge]  # noqa: E501
        edge = super(GMLAN_IDOEnumerator, self).get_new_edge(socket, config)
        if edge:
            state, new_state = edge
            new_state.tp = 1  # type: ignore
            return state, new_state
        return None

    @staticmethod
    def enter_state_with_tp(sock, conf, kwargs):
        # type: (_SocketUnion, AutomotiveTestCaseExecutorConfiguration, Dict[str, Any]) -> bool  # noqa: E501
        GMLAN_TPEnumerator.enter(sock, conf, kwargs)
        if GMLAN_IDOEnumerator.enter_diagnostic_session(sock):
            return True
        else:
            GMLAN_TPEnumerator.cleanup(sock, conf)
            return False

    def get_transition_function(self, socket, edge):
        # type: (_SocketUnion, _Edge) -> Optional[_TransitionTuple]
        return self.enter_state_with_tp, {"desc": "IDO_TP"}, GMLAN_TPEnumerator.cleanup  # noqa: E501

    def _get_table_entry_y(self, tup):
        # type: (_AutomotiveTestCaseScanResult) -> str
        return "InitiateDiagnosticOperation:"


class GMLAN_RDBIEnumerator(GMLAN_Enumerator):
    _description = "Readable data identifier per state"

    def _get_initial_requests(self, **kwargs):
        # type: (Any) -> Iterable[Packet]
        scan_range = kwargs.pop("scan_range", range(0x100))
        return (GMLAN() / GMLAN_RDBI(dataIdentifier=x) for x in scan_range)

    @staticmethod
    def print_information(resp):
        # type: (Packet) -> str
        load = bytes(resp)[2:] if len(resp) > 3 else b"No data available"
        return "PR: %r" % ((load[:17] + b"...") if len(load) > 20 else load)

    def _get_table_entry_y(self, tup):
        # type: (_AutomotiveTestCaseScanResult) -> str
        return "0x%04x: %s" % (tup[1].dataIdentifier,
                               tup[1].sprintf("%GMLAN_RDBI.dataIdentifier%"))

    def _get_table_entry_z(self, tup):
        # type: (_AutomotiveTestCaseScanResult) -> str
        return self._get_label(tup[2], self.print_information)


class GMLAN_WDBIEnumerator(GMLAN_Enumerator):
    _description = "Writeable data identifier per state"
    _supported_kwargs = copy.copy(GMLAN_Enumerator._supported_kwargs)
    _supported_kwargs.update({
        'rdbi_enumerator': (GMLAN_RDBIEnumerator, None)
    })

    _supported_kwargs_doc = ServiceEnumerator._supported_kwargs_doc + """
        :param rdbi_enumerator: Specifies an instance of a GMLAN_RDBIEnumerator
                                which is used to extract possible data
                                identifiers.
        :type rdbi_enumerator: GMLAN_RDBIEnumerator"""

    def execute(self, socket, state, **kwargs):
        # type: (_SocketUnion, EcuState, Any) -> None
        super(GMLAN_WDBIEnumerator, self).execute(socket, state, **kwargs)

    execute.__doc__ = _supported_kwargs_doc

    def _get_initial_requests(self, **kwargs):
        # type: (Any) -> Iterable[Packet]
        scan_range = kwargs.pop("scan_range", range(0x100))
        rdbi_enumerator = kwargs.pop("rdbi_enumerator", None)
        if rdbi_enumerator is None:
            return (GMLAN() / GMLAN_WDBI(dataIdentifier=x) for x in scan_range)
        elif isinstance(rdbi_enumerator, GMLAN_RDBIEnumerator):
            return (GMLAN() / GMLAN_WDBI(dataIdentifier=t.resp.dataIdentifier,
                                         dataRecord=bytes(t.resp)[2:])
                    for t in rdbi_enumerator.filtered_results
                    if t.resp.service != 0x7f and len(bytes(t.resp)) >= 2)
        else:
            raise Scapy_Exception("rdbi_enumerator has to be an instance "
                                  "of GMLAN_RDBIEnumerator")

    def _get_table_entry_y(self, tup):
        # type: (_AutomotiveTestCaseScanResult) -> str
        return "0x%02x: %s" % (tup[1].dataIdentifier,
                               tup[1].sprintf("%GMLAN_WDBI.dataIdentifier%"))

    def _get_table_entry_z(self, tup):
        # type: (_AutomotiveTestCaseScanResult) -> str
        return self._get_label(tup[2], "PR: Writeable")


class GMLAN_WDBISelectiveEnumerator(StagedAutomotiveTestCase):
    @staticmethod
    def __connector_rdbi_to_wdbi(rdbi, _):
        # type: (AutomotiveTestCaseABC, AutomotiveTestCaseABC) -> Dict[str, Any]  # noqa: E501
        return {"rdbi_enumerator": rdbi}

    def __init__(self):
        # type: () -> None
        super(GMLAN_WDBISelectiveEnumerator, self).__init__(
            [GMLAN_RDBIEnumerator(), GMLAN_WDBIEnumerator()],
            [None, self.__connector_rdbi_to_wdbi])


class GMLAN_SAEnumerator(GMLAN_Enumerator, StateGenerator):
    _description = "SecurityAccess supported"
    _transition_function_args = dict()  # type: Dict[_Edge, Tuple[int, Optional[Callable[[int], int]]]]  # noqa: E501
    _supported_kwargs = copy.copy(GMLAN_Enumerator._supported_kwargs)
    _supported_kwargs.update({
        'keyfunction': (None, None)
    })

    _supported_kwargs_doc = ServiceEnumerator._supported_kwargs_doc + """
        :param keyfunction: Specifies a function to generate the key from a
                            given seed.
        :type keyfunction: Callable[[int], int]"""

    def execute(self, socket, state, **kwargs):
        # type: (_SocketUnion, EcuState, Any) -> None
        super(GMLAN_SAEnumerator, self).execute(socket, state, **kwargs)

    execute.__doc__ = _supported_kwargs_doc

    def _get_initial_requests(self, **kwargs):
        # type: (Any) -> Iterable[Packet]
        scan_range = kwargs.pop("scan_range", range(1, 10, 2))
        return (GMLAN() / GMLAN_SA(subfunction=x) for x in scan_range)

    def _get_table_entry_y(self, tup):
        # type: (_AutomotiveTestCaseScanResult) -> str
        return "Subfunction %02d" % tup[1].subfunction

    def _get_table_entry_z(self, tup):
        # type: (_AutomotiveTestCaseScanResult) -> str
        return self._get_label(tup[2], lambda r: "PR: %s" % r.securitySeed)

    def pre_execute(self, socket, state, global_configuration):
        # type: (_SocketUnion, EcuState, AutomotiveTestCaseExecutorConfiguration) -> None  # noqa: E501
        if cast(ServiceEnumerator, self)._retry_pkt[state] and \
                not global_configuration.unittest:
            # this is a retry execute. Wait much longer than usual because
            # a required time delay not expired could have been received
            # on the previous attempt
            time.sleep(11)

    def _evaluate_retry(self,
                        state,  # type: EcuState
                        request,  # type: Packet
                        response,  # type: Packet
                        **kwargs  # type: Optional[Dict[str, Any]]
                        ):  # type: (...) -> bool

        if super(GMLAN_SAEnumerator, self)._evaluate_retry(
                state, request, response, **kwargs):
            return True

        if response.service == 0x7f and \
                self._get_negative_response_code(response) in [0x22, 0x37]:
            log_automotive.debug(
                "Retry %s because requiredTimeDelayNotExpired or "
                "requestSequenceError received",
                repr(request))
            return super(GMLAN_SAEnumerator, self)._populate_retry(
                state, request)
        return False

    def _evaluate_response(self,
                           state,  # type: EcuState
                           request,  # type: Packet
                           response,  # type: Optional[Packet]
                           **kwargs  # type: Optional[Dict[str, Any]]
                           ):  # type: (...) -> bool
        if super(GMLAN_SAEnumerator, self)._evaluate_response(
                state, request, response, **kwargs):
            return True

        if response is not None and \
                response.service == 0x67 and response.subfunction % 2 == 1:
            log_automotive.debug("Seed received. Leave scan to try a key")
            return True
        return False

    @staticmethod
    def get_seed_pkt(sock, level=1):
        # type: (_SocketUnion, int) -> Optional[Packet]
        req = GMLAN() / GMLAN_SA(subfunction=level)
        for _ in range(10):
            seed = sock.sr1(req, timeout=5, verbose=False)
            if seed is None:
                return None
            elif seed.service == 0x7f and \
                    GMLAN_Enumerator._get_negative_response_code(seed) != 0x37:
                log_automotive.info(
                    "Security access no seed! NR: %s", repr(seed))
                return None

            elif seed.service == 0x7f and \
                    GMLAN_Enumerator._get_negative_response_code(seed) == 0x37:
                log_automotive.info("Security access retry to get seed")
                time.sleep(10)
                continue
            else:
                return seed
        return None

    @staticmethod
    def evaluate_security_access_response(res, seed, key):
        # type: (Optional[Packet], Packet, Optional[Packet]) -> bool
        if res is None or res.service == 0x7f:
            log_automotive.debug(repr(seed))
            log_automotive.debug(repr(key))
            log_automotive.debug(repr(res))
            log_automotive.info("Security access error!")
            return False
        else:
            log_automotive.info("Security access granted!")
            return True

    @staticmethod
    def get_key_pkt(seed, keyfunction, level=1):
        # type: (Packet, Callable[[int], int], int) -> Optional[Packet]
        try:
            s = seed.securitySeed
        except AttributeError:
            return None

        return cast(Packet, GMLAN() / GMLAN_SA(subfunction=level + 1,
                                               securityKey=keyfunction(s)))

    @staticmethod
    def get_security_access(sock, level=1, seed_pkt=None, keyfunction=None):
        # type: (_SocketUnion, int, Optional[Packet], Optional[Callable[[int], int]]) -> bool  # noqa: E501
        log_automotive.info(
            "Try bootloader security access for level %d" % level)
        if seed_pkt is None:
            seed_pkt = GMLAN_SAEnumerator.get_seed_pkt(sock, level)
            if not seed_pkt:
                return False

        if keyfunction is None:
            return False

        key_pkt = GMLAN_SAEnumerator.get_key_pkt(seed_pkt, keyfunction, level)
        if key_pkt is None:
            return False

        res = sock.sr1(key_pkt, timeout=5, verbose=False)
        return GMLAN_SAEnumerator.evaluate_security_access_response(
            res, seed_pkt, key_pkt)

    @staticmethod
    def transition_function(sock, _, kwargs):
        # type: (_SocketUnion, AutomotiveTestCaseExecutorConfiguration, Dict[str, Any]) -> bool  # noqa: E501
        return GMLAN_SAEnumerator.get_security_access(
            sock, level=kwargs["sec_level"], keyfunction=kwargs["keyfunction"])

    def get_new_edge(self, socket, config):
        # type: (_SocketUnion, AutomotiveTestCaseExecutorConfiguration) -> Optional[_Edge]  # noqa: E501
        last_resp = self._results[-1].resp
        last_state = self._results[-1].state

        if last_resp is None or last_resp.service == 0x7f:
            return None

        try:
            if last_resp.service != 0x67 or \
                    last_resp.subfunction % 2 != 1:
                return None

            seed = last_resp
            sec_lvl = seed.subfunction
            kf = config[self.__class__.__name__].get("keyfunction", None)

            if self.get_security_access(socket, level=sec_lvl,
                                        seed_pkt=seed, keyfunction=kf):
                log_automotive.debug("Security Access found.")
                # create edge
                new_state = copy.copy(last_state)
                new_state.security_level = seed.subfunction + 1  # type: ignore  # noqa: E501
                if last_state == new_state:
                    return None
                edge = (last_state, new_state)
                self._transition_function_args[edge] = (sec_lvl, kf)
                return edge
        except AttributeError:
            pass

        return None

    def get_transition_function(self, socket, edge):
        # type: (_SocketUnion, _Edge) -> Optional[_TransitionTuple]
        return self.transition_function, {
            "sec_level": self._transition_function_args[edge][0],
            "keyfunction": self._transition_function_args[edge][1],
            "desc": "SA=%d" % self._transition_function_args[edge][0]}, None


class GMLAN_RDEnumerator(GMLAN_Enumerator, StateGeneratingServiceEnumerator):
    _description = "RequestDownload supported"

    def _get_initial_requests(self, **kwargs):
        # type: (Any) -> Iterable[Packet]
        return [GMLAN() / GMLAN_RD(memorySize=0x10)]

    def _get_table_entry_y(self, tup):
        # type: (_AutomotiveTestCaseScanResult) -> str
        return "RequestDownload:"


class GMLAN_PMEnumerator(GMLAN_Enumerator, StateGeneratingServiceEnumerator):
    _description = "ProgrammingMode supported"

    def _get_initial_requests(self, **kwargs):
        # type: (Any) -> Iterable[Packet]
        raise NotImplementedError()

    def execute(self, socket, state, timeout=1, execution_time=1200, **kwargs):
        # type: (_SocketUnion, EcuState, int, int, Any) -> None
        supported = GMLAN_InitDiagnostics(
            cast(SuperSocket, socket), timeout=20,
            unittest=kwargs.get("unittest", False))
        # TODO: Refactor result storage
        if supported:
            self._store_result(
                state, GMLAN() / GMLAN_PM(), GMLAN(service=0xE5))
        else:
            self._store_result(
                state, GMLAN() / GMLAN_PM(),
                GMLAN() / GMLAN_NR(returnCode=0x11, requestServiceId=0xA5))

        self._state_completed[state] = True

    def get_new_edge(self, socket, config):
        # type: (_SocketUnion, AutomotiveTestCaseExecutorConfiguration) -> Optional[_Edge]  # noqa: E501
        edge = super(GMLAN_PMEnumerator, self).get_new_edge(socket, config)
        if edge:
            state, new_state = edge
            new_state.tp = 1  # type: ignore
            new_state.communication_control = 1  # type: ignore
            return state, new_state
        return None

    @staticmethod
    def enter_state_with_tp(sock, conf, kwargs):
        # type: (_SocketUnion, AutomotiveTestCaseExecutorConfiguration, Dict[str, Any]) -> bool  # noqa: E501
        GMLAN_TPEnumerator.enter(sock, conf, kwargs)
        res = GMLAN_InitDiagnostics(cast(SuperSocket, sock), timeout=20,
                                    unittest=conf.unittest)
        if not res:
            GMLAN_TPEnumerator.cleanup(sock, conf)
            return False
        else:
            return True

    def get_transition_function(self, socket, edge):
        # type: (_SocketUnion, _Edge) -> Optional[_TransitionTuple]
        return self.enter_state_with_tp, {"desc": "PM_TP"}, \
            GMLAN_TPEnumerator.cleanup

    def _get_table_entry_y(self, tup):
        # type: (_AutomotiveTestCaseScanResult) -> str
        return "ProgrammingMode:"


class GMLAN_RDBPIEnumerator(GMLAN_Enumerator):
    _description = "Readable parameter identifier per state"

    def _get_initial_requests(self, **kwargs):
        # type: (Any) -> Iterable[Packet]
        scan_range = kwargs.pop("scan_range", range(0x10000))
        return (GMLAN() / GMLAN_RDBPI(identifiers=[x]) for x in scan_range)

    def _get_table_entry_y(self, tup):
        # type: (_AutomotiveTestCaseScanResult) -> str
        return "0x%04x: %s" % (
            tup[1].identifiers[0],
            tup[1].sprintf("%GMLAN_RDBPI.identifiers%")[1:-1])

    def _get_table_entry_z(self, tup):
        # type: (_AutomotiveTestCaseScanResult) -> str
        return self._get_label(tup[2], GMLAN_RDBIEnumerator.print_information)


class GMLAN_RMBAEnumerator(GMLAN_Enumerator):
    _description = "Readable Memory Addresses and negative response per state"

    _supported_kwargs = copy.copy(GMLAN_Enumerator._supported_kwargs)
    _supported_kwargs.update({
        'probe_width': (int, lambda x: x >= 0),
        'random_probes_len': (int, lambda x: x >= 0),
        'sequential_probes_len': (int, lambda x: x >= 0)
    })

    _supported_kwargs_doc = GMLAN_Enumerator._supported_kwargs_doc + """
        :param int probe_width: Memory size of a probe.
        :param int random_probes_len: Number of probes.
        :param int sequential_probes_len: Size of a memory block during
                                          sequential probing."""

    def execute(self, socket, state, **kwargs):
        # type: (_SocketUnion, EcuState, Any) -> None
        super(GMLAN_RMBAEnumerator, self).execute(socket, state, **kwargs)

    execute.__doc__ = _supported_kwargs_doc

    def __init__(self):
        # type: () -> None
        super(GMLAN_RMBAEnumerator, self).__init__()
        self.random_probe_finished = defaultdict(bool)  # type: Dict[EcuState, bool]  # noqa: E501
        self.points_of_interest = defaultdict(list)  # type: Dict[EcuState, List[Tuple[int, bool]]]  # noqa: E501
        self.probe_width = 0x10  # defines the memorySize of a request
        self.highest_possible_addr = \
            2 ** (8 * conf.contribs['GMLAN']['GMLAN_ECU_AddressingScheme']) - 1
        self.random_probes_len = \
            min(10 ** conf.contribs['GMLAN']['GMLAN_ECU_AddressingScheme'],
                0x5000)
        self.sequential_probes_len = \
            10 ** (conf.contribs['GMLAN']['GMLAN_ECU_AddressingScheme'])

    def _get_initial_requests(self, **kwargs):
        # type: (Any) -> Iterable[Packet]
        self.probe_width = kwargs.pop("probe_width", self.probe_width)
        self.random_probes_len = \
            kwargs.pop("random_probes_len", self.random_probes_len)
        self.sequential_probes_len = \
            kwargs.pop("sequential_probes_len", self.sequential_probes_len)
        addresses = random.sample(
            range(0, self.highest_possible_addr, self.probe_width),
            self.random_probes_len)
        scan_range = kwargs.pop("scan_range", addresses)
        return (GMLAN() / GMLAN_RMBA(memoryAddress=x,
                                     memorySize=self.probe_width)
                for x in scan_range)

    def post_execute(self, socket, state, global_configuration):
        # type: (_SocketUnion, EcuState, AutomotiveTestCaseExecutorConfiguration) -> None  # noqa: E501
        if not self._state_completed[state]:
            return

        if not self.random_probe_finished[state]:
            log_automotive.info("Random memory probing finished")
            self.random_probe_finished[state] = True
            for tup in [t for t in self.results_with_positive_response
                        if t.state == state]:
                self.points_of_interest[state].append(
                    (tup.req.memoryAddress, True))
                self.points_of_interest[state].append(
                    (tup.req.memoryAddress, False))

        if not len(self.points_of_interest[state]):
            return

        log_automotive.info(
            "Create %d memory points for sequential probing" %
            len(self.points_of_interest[state]))

        tested_addrs = [tup.req.memoryAddress for tup in self.results]
        pos_addrs = [tup.req.memoryAddress for tup in
                     self.results_with_positive_response if tup.state == state]

        new_requests = list()
        new_points_of_interest = list()

        for poi, upward in self.points_of_interest[state]:
            if poi not in pos_addrs:
                continue
            temp_new_requests = list()
            for i in range(
                    self.probe_width,
                    self.sequential_probes_len + self.probe_width,
                    self.probe_width):
                if upward:
                    new_addr = min(poi + i, self.highest_possible_addr)
                else:
                    new_addr = max(poi - i, 0)

                if new_addr not in tested_addrs:
                    pkt = GMLAN() / GMLAN_RMBA(memoryAddress=new_addr,
                                               memorySize=self.probe_width)
                    temp_new_requests.append(pkt)

            if len(temp_new_requests):
                new_points_of_interest.append(
                    (temp_new_requests[-1].memoryAddress, upward))
                new_requests += temp_new_requests

        self.points_of_interest[state] = list()

        if len(new_requests):
            self._state_completed[state] = False
            self._request_iterators[state] = new_requests
            self.points_of_interest[state] = new_points_of_interest
            log_automotive.info(
                "Created %d pkts for sequential probing" %
                len(new_requests))

    def show(self, dump=False, filtered=True, verbose=False):
        # type: (bool, bool, bool) -> Optional[str]
        s = super(GMLAN_RMBAEnumerator, self).show(dump, filtered, verbose)
        try:
            from intelhex import IntelHex

            ih = IntelHex()
            for tup in self.results_with_positive_response:
                for i, b in enumerate(tup.resp.dataRecord):
                    ih[tup.req.memoryAddress + i] = orb(b)

            ih.tofile("RMBA_dump.hex", format="hex")
        except ImportError:
            log_automotive.warning(
                "Install 'intelhex' to create a hex file of the memory")

        if dump and s is not None:
            return s + "\n"
        else:
            print(s)
            return None

    def _get_table_entry_y(self, tup):
        # type: (_AutomotiveTestCaseScanResult) -> str
        return "0x%04x" % tup[1].memoryAddress

    def _get_table_entry_z(self, tup):
        # type: (_AutomotiveTestCaseScanResult) -> str
        return self._get_label(tup[2], lambda r: "PR: %s" % r.dataRecord)


class GMLAN_TDEnumerator(GMLAN_Enumerator):
    _description = "Transfer Data support and negative response per state"

    def _get_initial_requests(self, **kwargs):
        # type: (Any) -> Iterable[Packet]
        scan_range = kwargs.pop("scan_range", range(0x1ff))
        temp = conf.contribs["GMLAN"]['GMLAN_ECU_AddressingScheme']
        # Shift operations to eliminate addresses not aligned to 4
        max_addr = (2 ** (temp * 8) - 1) >> 2
        addresses = (random.randint(0, max_addr) << 2 for _ in scan_range)
        return (GMLAN() / GMLAN_TD(subfunction=0, startingAddress=x)
                for x in addresses)

    def _get_table_entry_y(self, tup):
        # type: (_AutomotiveTestCaseScanResult) -> str
        return "0x%04x" % tup[1].startingAddress


class GMLAN_DCEnumerator(GMLAN_Enumerator):
    _description = "DeviceControl supported per state"

    def _get_initial_requests(self, **kwargs):
        # type: (Any) -> Iterable[Packet]
        scan_range = kwargs.pop("scan_range", range(0x100))
        return (GMLAN() / GMLAN_DC(CPIDNumber=x) for x in scan_range)

    def _get_table_entry_y(self, tup):
        # type: (_AutomotiveTestCaseScanResult) -> str
        return "0x%02x: %s" % \
               (tup[1].CPIDNumber, tup[1].sprintf("%GMLAN_DC.CPIDNumber%"))


# ########################## GMLAN SCANNER ###################################

class GMLAN_Scanner(AutomotiveTestCaseExecutor):
    @property
    def default_test_case_clss(self):
        # type: () -> List[Type[AutomotiveTestCaseABC]]
        return [GMLAN_ServiceEnumerator, GMLAN_TPEnumerator,
                GMLAN_IDOEnumerator, GMLAN_PMEnumerator,
                GMLAN_RDEnumerator, GMLAN_SAEnumerator, GMLAN_TDEnumerator,
                GMLAN_RMBAEnumerator,
                GMLAN_WDBISelectiveEnumerator, GMLAN_DCEnumerator]
